# =========================================================================
#
# interface to GCI
#  - initialize (to do)
#  - translating event in a version in which the PCO is identified;
#  - build a term representation that can be passed via GCI
#  - send a term via GCI
#  - receive a term from GCI
#

proc TorxGciSend {pco message} {
	if {[main_using_gci]} {
		GciSend $pco $message
	} else {
		send -i [pco2sid $pco] $message
	}
}

proc TorxGciStartTimer {timer value units} {
	if {[main_using_gci]} {
		GciStartTimer $timer $value $units
	} else {
		global timeout
		set timeout $value
	}
}

proc TorxGciReadTimer {timer} {
	if {[main_using_gci]} {
		return [GciReadTimer $timer]
	} else {
		return -1
	}
}

proc TorxGciCancelTimer {timer} {
	if {[main_using_gci]} {
		GciCancelTimer $timer
	} else {
		global timeout maininfo
		set timeout $maininfo(timeout)
	}
}

#
# after calling GciSnapshot global variable gciqueue will contain
# at least one element.
# Elements in gciqueue are one of the two following:
#  - a list of two elements: { output   output-value-in-term-format }
#  - a singleton list:       { delta }
#
proc TorxGciSnapshot {} {
	if {[main_using_gci]} {
		GciSnapshot
	} else {
		# NOTE: we need 'global sidinfo' because of the 'eval' below!!!
		global iutinfo gciqueue sidinfo
		eval $iutinfo(expect_command)
		# expect_command should call GciReceive or GciTimeout instead
		# of just setting the 'result' variable
		# Also: the logging routines should call GciLog
		set gciresult [result2gcivalue $result]
		lappend gciqueue $gciresult
	}
}

# =========================================================================

proc TorxGciInit {} {
	global pcoinfo
	if {[main_using_gci]} {
		foreach pco $pcoinfo(pcos) {
			buildPcoNameMapping $pco
		}
		buildPcoNameMapping TorXAnyGeneric
		GciInit
	} else {
		# nothing to do
	}
}

proc TorxGciGenericPco {} {
	if {[main_using_gci]} {
		return [GciGetGenericPco]
	} else {
		# nothing to do
	}
}

# =========================================================================

proc result2gcivalue {result} {
	if {[main_using_gci]} {
		# nothing to do
	} else {
		return $result
	}
}

proc event2gcivalue {g exp} {
	if {[main_using_gci]} {
		return [list nodes [event2term [join [concat $g $exp] " ! "]]]
	} else {
		return [encode_event $g $exp]
	}
}

proc gcivalue2event {v} {
	if {[main_using_gci]} {
		set eventlist  [term2event $v]
		set result [list [lindex $eventlist 0] [lrange  $eventlist 1 end]]
		return $result
	} else {
		return $v
	}
}

# =========================================================================
