proc Debug_Stack { fn args } {
	global debug
	global errorInfo
	if {([info exists debug(Stack)] && $debug(Stack)) ||
	    ([info exists debug(Stack,$fn)] && $debug(Stack,$fn)) } {
		catch {puts stderr "$fn stack: ==============="}
		foreach l [split $errorInfo \n] {
			catch {puts stderr "Stack: $l"}
		}
		catch {puts stderr "$fn end stack ==============="}
		catch {puts stderr "$fn levelinfo: ==============="}
	    set level [info level]
	    for {set l 1} {$l <= $level} {incr l} {
			catch {puts stderr [info level $l]}
	    }
		catch {puts stderr "$fn end levelinfo ==============="}
	}
}


proc Debug_FnArgs { fn args } {
	global debug
	if {([info exists debug(FnArgs)] && $debug(FnArgs)) ||
	    ([info exists debug(FnArgs,$fn)] && $debug(FnArgs,$fn)) } {
		catch {puts stderr "$fn [llength $args]: ==============="}
		foreach e $args {
			if {[string length $e] > 100} {
				catch {puts stderr "	[string range $e 0 100] ..."}
			} else {
				catch {puts stderr "	$e"}
			}
		}
		catch {puts stderr "$fn end args ==============="}
	}
}
proc Debug_FnRes { fn args } {
	global debug
	if {([info exists debug(FnRes)] && $debug(FnRes)) ||
	    ([info exists debug(FnRes,$fn)] && $debug(FnRes,$fn)) } {
		if {[llength $args] > 0} {
				catch {puts stderr "$fn result: $args"}
		} else {
				catch {puts stderr "$fn no result (void)"}
		}
	}
}

# currently we specifically pass the names of the global variables
# that have to be set; in the future: do we want to use uplevel instead?
proc Debug_FnGlobals { fn msg globalDecls args } {
	global debug
	if {([info exists debug(FnGlobals)] && $debug(FnGlobals)) ||
	    ([info exists debug(FnGlobals,$fn)] && $debug(FnGlobals,$fn)) } {
	    set txt "$fn $msg"
		foreach d $globalDecls {
			global $d
		}
		foreach varName $args {
			if {[info exists $varName]} {
				append txt " $varName=[set $varName]"
			} else {
				append txt " $varName (unset)"
			}
		}
		catch {puts stderr $txt}
	}
}
proc Debug_Tags { txt w } {
	global debug
	if {[info exists debug(Tags)] && $debug(Tags)} {
		catch {puts stderr "$txt"}
		if {[winfo exists $w]} {
			foreach t [$w tag names] {
				catch {puts stderr "	tag $t [$w tag ranges $t]"}
			}
		} else {
			catch {puts stderr "Debug_Tags error: no window $w"}
		}
	}
}
proc Debug_Marks { txt w } {
	global debug
	if {[info exists debug(Marks)] && $debug(Marks)} {
		catch {puts stderr "$txt"}
		if {[winfo exists $w]} {
			foreach t [$w mark names] {
				catch {puts stderr "	mark $t [$w index $t]"}
			}
		} else {
			catch {puts stderr "Debug_Marks error: no window $w"}
		}
	}
}
proc Debug_Msg { fn txt } {
	global debug
	if {([info exists debug(Msg)] && $debug(Msg)) ||
	    ([info exists debug(Msg,$fn)] && $debug(Msg,$fn)) } {
		if {![info exists debug(fn,disabled,$fn)]} {
			catch {puts stderr "$fn $txt"}
		}
	}
}


