# =========================================================================
#

proc setup_stderr_patterns {} {
	global sidinfo
 	set sidinfo(bg) {}
 
#		-i sidinfo(bg) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
#			global sidinfo
#			set txt $expect_out(1,string)
#			send_error "$sidinfo(sid,name,$expect_out(spawn_id)) Debug: $txt\n"
#		}

	expect_background {
		-i sidinfo(bg) full_buffer {
			global sidinfo
			warning "SPEC: Buffer Overflow!!!\n"
			set txt $expect_out(buffer)
			send_error "$sidinfo(sid,name,$expect_out(spawn_id)): Stderr Buffer Contents Skipped: $txt\n"
			exit
		}
		-i sidinfo(bg) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
			global sidinfo
			set txt $expect_out(1,string)
			send_error "$sidinfo(sid,name,$expect_out(spawn_id)) Debug: $txt\n"
		}
		-i sidinfo(bg) -re "(\[^\r\n]*)\r*\n" {
			global sidinfo
			set txt $expect_out(1,string)
			send_error "$sidinfo(sid,name,$expect_out(spawn_id)) Stderr: $txt\n"
		}
	}
}

# =========================================================================
#
# main loop end-of-test routine
#

proc main_go_on {} {
	global maininfo
	if { $maininfo(stopnow) } {
		return 0
	}
	if { $maininfo(endofspec) ||
	  (($maininfo(nrofstepsmax) > 0) &&
           ($maininfo(nrofstepsdone) >= $maininfo(nrofstepsmax))) } {
        if {[string compare $maininfo(verdict) inconclusive] == 0} {
            set maininfo(verdict) pass
        }
		return 0
	}
	return 1
}

# =========================================================================
#
# primitives used in the main loop
#

proc spec_test_eot {} {
	set result ""
	io_send_command spec EOT
	io_receive_answer spec {
		-singleline {
			EOT_YES {} {
				set result [list yes]
			}
			EOT_NO {} {
				set result [list no]
			}
		}
	}
	debugLvl 2 "spec_test_eot result ($result)\n"
	return $result
}

# NOTE: spec_get_iokind still recognizes the 'old' A_IO_*
# answers from the Primer, for pseudo backward compatibility.
proc spec_get_iokind {sug} {
	set sug_iokind [lindex $sug 0]
	set sug_channel [lindex $sug 1]
	set iosug {}
	foreach name {iokind channel} {
		if {[string compare [set sug_[set name]] ""] != 0} {
			lappend iosug [list [set name] [set sug_[set name]]]
		}
	}
	set result ""
	io_send_command spec IOKIND $iosug
	io_receive_answer spec {
		-singleline {
			IOKIND {channel iokind} {
				set result [list $iokind $channel]
			}
			IO_INPUT {} {
				set result [list input [default_channel input]]
			}
			IO_OUTPUT {} {
				set result [list output [default_channel output]]
			}
			"ERROR Inconsistency:" {} {
				set result [list error $line]
			}
		}
	}
	debugLvl 2 "spec_get_iokind result ($result)\n"
	return $result
}

proc spec_get_input {chan sug pred_sug} {
	set ev $sug
	set tuples {}
	set result ""
	lappend tuples [list channel $chan]
	if {[string compare $ev ""] != 0} {
		# TODO? should remove the join here too, like we did for predicates below?
		lappend tuples [list event [join $ev " "]]
	}
	if {[string compare $pred_sug ""] != 0} {
		lappend tuples [list predicates $pred_sug]
	}
	io_send_command spec GETINPUT $tuples
	io_receive_answer spec {
		-singleline {
			GETINPUT {event channel suspension predicates} {
				set result [list input_ok $channel $event $suspension $predicates]
			}
			GETINPUT_OK {event channel suspension predicates} {
				set result [list input_ok $channel $event $suspension $predicates]
			}
			GETINPUT_ERROR {reason} {
				set result [list input_error $reason]
			}
			"ERROR Inconsistency" {} {
				set result [list input_error $line]
			}
		}
	}
	debugLvl 2 "spec_get_input result ($result)\n"
	return $result
}

#
# spec_choose_from_menu
# This routine takes a suggestion as argument.
# If the suggestion is the empty event, and $maininfo(chooseInputs) is set,
# it asks the specification module for the list of input events, and
# chooses one element from this list (using a random number generator).
# We will then invoke spec_get_input with this chosen input event.
#
# So, when $maininfo(chooseInputs) is set, it is _not_ the spec module
# who will randomly choose an event, but this routine will do this.
#
# Why do we want this behaviour?
# We want this, because, for promela, we want our instantiator module
# to be able to instantiate input events. The instantiator can only
# do that if the C_INPUT command sent to the spec module contains the
# complete suggested event (taken from the input menu).
#
proc spec_choose_from_menu {sug preds} {
	global sidinfo maininfo specinfo
	debugLvl 2 "spec_choose_from_menu ($sug)\n"
	set result $sug
	if {([llength [lindex $sug 2]] == 0) && ($maininfo(chooseInputs))} {
		set kw_inputs [spec_get_inputs]
		set inputs [lindex $kw_inputs 1]
		debugLvl 2 "spec_choose_from_menu inputs ($inputs)\n"
		set nr_of_el [llength $inputs]
		set idx [RandomRange $nr_of_el]
		set el [lindex $inputs $idx]
		set el_chan [lindex $el 0]
		set el_ev [lindex $el 1]
		set el_susp [lindex $el 2]
		set el_preds [lindex $el 3]
		set result [list input $el_chan $el_ev $el_susp]
		debugLvl 2 "spec_choose_from_menu (nr_of_el: $nr_of_el)(idx: $idx) result ($result)\n"
	}
	debugLvl 2 "spec_choose_from_menu result ($result)\n"
	return $result
}

# kind should be INPUT or OUTPUT
proc spec_get_events {kind} {
	set result ""
	set events {}
	switch -- $kind {
		INPUT { set type inputs }
		OUTPUT { set type outputs }
		default { return $result }
	}
	io_send_command spec "[set kind]S"
	io_receive_answer spec {
		-multiline "[set kind]S" {
			EVENT {event channel suspension predicates} {
				lappend events [list $channel $event $suspension $predicates]
			}
			end {
				set result [list $type $events ]
			}
		}
	}
	debugLvl 2 "spec_get_events result ($result)\n"
	return $result
}

proc spec_get_outputs {} {
	return [spec_get_events OUTPUT]
}

proc spec_get_inputs {} {
	return [spec_get_events INPUT]
}

proc spec_get_state {} {
	set result ""
	set state ""
	io_send_command spec STATE
	io_receive_answer spec {
		-multiline STATE {
			end {set result [list state $state ]}
			{} {} {
				lappend state $line
			}
		}
	}
	debugLvl 2 "spec_get_state result ($result)\n"
	return $result
}

proc spec_get_stats {} {
	io_send_command spec STATS
	set result ""
	set stats ""
	io_receive_answer spec {
		-multiline STATS {
			STATS {} {
				lappend stats $line
			}
			end {set result [list stats $stats ]}
		}
		-singleline {
			STATS {} {set result [list stats [list $line] ]}
		}
	}
	debugLvl 2 "spec_get_stats result ($result)\n"
	return $result
}

proc get_stateid {c} {
	io_send_command $c STATEID
	set result ""
	set stats ""
	io_receive_answer $c {
		-multiline STATEID {
			STATEID {identity super init trans} {
				if {[string compare $trans ""] == 0} {
					lappend stats [list $identity super $super init $init]
				} else {
					lappend stats [list $identity super $super init $init trans $trans]
				}
			}
			end {set result [list stateid $stats ]}
		}
		-singleline {
			STATEID {super init trans} {
				if {[string compare $trans ""] == 0} {
					lappend stats [list super $super init $init]
				} else {
					lappend stats [list super $super init $init trans $trans]
				}
				set result [list stateid $stats ]
			}
		}
	}
	debugLvl 2 "get_stateid result ($result)\n"
	return $result
}

proc set_stateid {c s} {
	set tuples {}
	lappend tuples [list super $s]
	io_send_command $c GOTO $tuples
	set result ""
	io_receive_answer $c {
		-singleline {
			GOTO_OK {} {
				set result [list goto]
			}
			GOTO_ERROR {} {
				set result [list goto_error $line]
			}
			"ERROR UnknownCommand:" {} {
				warning "spec does not implement GOTO command: $line"
				set result [list goto_error "spec does not implement GOTO command: $line"]
			}
		}
	}
	debugLvl 2 "set_stateid result ($result)\n"
	return $result
}

proc spec_get_config {} {
	set result ""
	set config ""
	io_send_command spec GETCONFIG
	io_receive_answer spec {
		-multiline GETCONFIG {
			end {set result [list config $config ]}
			CONFIG {} {
				lappend config $line
			}
		}
	}
	debugLvl 2 "spec_get_config result ($result)\n"
	return $result
}

# get ikind result from adaptor
# if result is 'incomplete', then either return the sug, or,
# if sug and adapter-result seem compatible, return combination
#    of sug and adapter result.
proc adaptor_get_iokind {sug} {
	set sug_iokind [lindex $sug 0]
	set sug_channel [lindex $sug 1]
	set iosug {}
	foreach name {iokind channel} {
		if {[string compare [set sug_[set name]] ""] != 0} {
			lappend iosug [list [set name] [set sug_[set name]]]
		}
	}
	set result ""
	io_send_command adaptor IOKIND $iosug
	io_receive_answer adaptor {
		-singleline {
			IOKIND {channel iokind} {
				if {([string compare $channel ""] != 0) &&
				    ([string compare $iokind ""] != 0)} {
					set result [list $iokind $channel]
				} elseif {([string compare $channel ""] == 0) &&
				    ([string compare $iokind ""] == 0)} {
					set result [list $sug_iokind $sug_channel]
				} elseif {([string compare $channel ""] == 0) &&
				    ([string compare $iokind ""] != 0)} {
					if {([string compare $iokind $sug_iokind] == 0)} {
						set result [list $sug_iokind $sug_channel]
					} else {
						set result [list $iokind $channel]
					}
				} elseif {([string compare $channel ""] != 0) &&
				    ([string compare $iokind ""] == 0)} {
					if {([string compare $channel $sug_channel] == 0)} {
						set result [list $sug_iokind $sug_channel]
					} else {
						set result [list $iokind $channel]
					}
				} else {
					set result [list $iokind $channel]
				}
			}
			"ERROR UnknownCommand:" {} {
				warning "adaptor does not implement IOKIND command: $line"
				set result [list $sug_iokind $sug_channel]
			}
			"ERROR Inconsistency:" {} {
				set result [list error $line]
			}
		}
	}
	debugLvl 2 "adaptor_get_iokind result ($result)\n"
	return $result
}

proc adaptor_put_input { chan ev preds } {
	debugLvl 1 "adaptor_put_input: C_INPUT channel=$chan\tevent=$ev\tpredicates=$preds\n"
	io_send_command adaptor INPUT [list [list channel $chan] [list event $ev] [list predicates $preds]]
	set result [adaptor_waitfor_answer $chan input INPUT]
	# puts stderr "adaptor_put_input ($chan) ($ev) : ($result)"
	return $result
}

proc adaptor_get_output { chan } {
	debugLvl 1 "C_OUTPUT channel=$chan\n"
	io_send_command adaptor OUTPUT [list [list channel $chan]]
	set result [adaptor_waitfor_answer $chan output OUTPUT]
	# puts stderr "adaptor_get_output ($chan) : ($result)"
	return $result
}

proc adaptor_waitfor_answer { chan kind KIND } {
#	global io
#	set io(debug) 5
#	set io(debug,expect) 5
	set result ""
	io_receive_answer adaptor {
		-singleline {
			"[set KIND]" {event suspension predicates pco channel timestamp concrete} {
				if {[string compare $suspension ""] == 0} {
					set suspension 0
				}
				set result [list [set kind]_ok $channel $event $suspension $predicates $pco $timestamp $concrete]
			}
			"[set KIND]_OK" {event suspension predicates pco channel timestamp concrete} {
				if {[string compare $suspension ""] == 0} {
					set suspension 0
				}
				set result [list [set kind]_ok $channel $event $suspension $predicates $pco $timestamp $concrete]
			}
			"[set KIND]_ERROR" {reason predicates explanation} {
				puts stderr "[set KIND]_ERROR $reason $predicates $explanation"
				if {[string compare $reason ""] != 0} {
					set result [list [set kind]_error $reason $predicates $explanation]
				} elseif {[info exists line]} {
					set result [list [set kind]_error "" "" $line]
				} else {
					set result [list [set kind]_error]
				}
			}
		}
	}
#	catch {unset io(debug)}
#	catch {unset  io(debug,expect)}
	debugLvl 2 "adaptor_waitfor_answer result ($result)\n"
	return $result
}

proc spec_check_output { chan ev susp preds } {
	debugLvl 1 "C_OUTPUT channel=$chan\tevent=$ev\tsuspension=$susp\tpredicates=$preds\n"
	set result ""
	set tuples {}
	lappend tuples [list channel $chan]
	lappend tuples [list event $ev]
	lappend tuples [list suspension $susp]
	if {[string compare $preds ""] != 0} {
		lappend tuples [list predicates $preds]
	}
	io_send_command spec OUTPUT $tuples
	io_receive_answer spec {
		-singleline {
			OUTPUT_OK {event channel suspension predicates} {
				set result [list output_ok $channel $event $suspension $predicates]
			}
			OUTPUT_ERROR {} {
				set result [list output_error $line]
			}
		}
	}
	debugLvl 2 "spec_check_output result ($result)\n"
	return $result
}

proc check_output { chan ev susp preds } {
	global maininfo
	set result [spec_check_output $chan $ev $susp $preds  ]
	switch -- [lindex $result 0] \
	  output_ok {
		# skip - everything is ok
	} output_error {
		log_time_and_state
		set maininfo(verdict) fail
		utlog_touser_output "spec disagrees with iut\n"
		utlog_touser_output "outputs expected by spec:\n"
		set outputs [spec_get_outputs]
		switch -- [lindex $outputs 0] \
		  outputs {
			utlog_touser_menu expected [lindex $outputs 1]
			utlog_touser_output "(end of expected outputs)\n"
			foreach o [lindex $outputs 1] {
				utlog_tofile_expectation "output" $o
			}
		} error {
			fatal "error from specification\n"
		}
		utlog_touser_output "state according to spec:\n"
		set state [spec_get_state]
		switch -- [lindex $state 0] \
		  state {
			utlog_touser_output [join [lindex $state 1] "\n"]
			utlog_touser_output "(end of state)\n"
		} error {
			fatal "error from specification\n"
		}
		clean_exit 0
	} error {
		fatal "error from specification\n"
	}
}

proc spec_check_input { chan ev susp preds } {
	# debugLvl 1 "C_INPUT channel=$chan\tevent=$ev\tsuspension=$susp\n"
	debugLvl 1 "C_INPUT channel=$chan\tevent=$ev\tpredicates=$preds\n"
	set result ""
	set tuples {}
	lappend tuples [list channel $chan]
	lappend tuples [list event $ev]
	lappend tuples [list suspension $susp]
	if {[string compare $preds ""] != 0} {
		lappend tuples [list predicates $preds]
	}
	io_send_command spec INPUT $tuples
#	io_send_command spec INPUT [list [list channel $chan] \
#	                           [list event $ev]  ]
	io_receive_answer spec {
		-singleline {
			INPUT {event channel suspension predicates} {
				set result [list input_ok $channel $event $suspension $predicates]
			}
			INPUT_OK {event channel suspension predicates} {
				set result [list input_ok $channel $event $suspension $predicates]
			}
			INPUT_ERROR {} {
				set result [list input_error $line]
			}
		}
	}
	debugLvl 2 "spec_check_input result ($result)\n"
	return $result
}

proc check_input { chan ev susp preds } {
	global maininfo
	set result [spec_check_input $chan $ev $susp $preds  ]
	switch -- [lindex $result 0] \
	  input_ok {
		# skip - everything is ok
	} input_error {
		log_time_and_state
		set maininfo(verdict) fail
		utlog_touser_output "spec disagrees with iut\n"
		utlog_touser_output "inputs expected by spec:\n"
		set inputs [spec_get_inputs]
		switch -- [lindex $inputs 0] \
		  inputs {
			utlog_touser_menu expected [lindex $inputs 1]
			utlog_touser_output "(end of expected inputs)\n"
			foreach o [lindex $inputs 1] {
				utlog_tofile_expectation "input" $o
			}
		} error {
			fatal "error from specification\n"
		}
		utlog_touser_output "state according to spec:\n"
		set state [spec_get_state]
		switch -- [lindex $state 0] \
		  state {
			utlog_touser_output [join [lindex $state 1] "\n"]
			utlog_touser_output "(end of state)\n"
		} error {
			fatal "error from specification\n"
		}
		clean_exit 0
	} error {
		fatal "error from specification\n"
	}
}
# =========================================================================

proc instantiate { preds } {
	global inst instantiator
	set result [list inst_ok $preds]
	if {[string compare $instantiator ""] != 0} {
		io_send_command $instantiator EVENT [list                  \
				[list predicates $preds]           \
				]
		io_receive_answer $instantiator {
			-singleline {
				EVENT {predicates} {
					debugLvl 1 "instantiate ($preds) -> $predicates"
					set result [list inst_ok $predicates]
				}
				EVENT_ERROR {reason predicates explanation} {
					set result [list inst_error $predicates $reason $explanation]
				}
				{} {} {
					set result [list inst_error "" ""  $line]
				}
			}
		}
		debugLvl 0  "instantiate resul=$result"
	} else {
		debugLvl 1 "instantiate no inst ($preds)"
	}
	return $result
}

# =========================================================================

proc instantiate_all { ev susp preds } {
	global inst instantiator
	set result [list inst_ok $ev $preds]
	if {[string compare $instantiator ""] != 0} {
		io_send_command $instantiator EVENT [list                  \
				[list event $ev]           \
				[list predicates $preds]           \
				]
		io_receive_answer $instantiator {
			-singleline {
				EVENT {event suspension predicates} {
					debugLvl 1 "instantiate ($ev)($susp)($preds) -> ($event)($suspension)($predicates)"
					set result [list inst_ok [list $event $suspension $predicates] ]
				}
				EVENT_ERROR {reason predicates explanation} {
					set result [list inst_error $predicates $reason $explanation]
				}
				{} {} {
					set result [list inst_error "" ""  $line]
				}
			}
			-multiline EVENT {
				EVENT {event suspension predicates} {
					debugLvl 1 "instantiate part ($ev)($susp)($preds) -> ($event)($suspension)($predicates)"
					lappend events [list $event $suspension $predicates]
				}
				end {
					set result [list inst_ok $events ]
				}
			}
		}
		debugLvl 0  "instantiate resul=$result"
	} else {
		debugLvl 1 "instantiate no inst ($preds)"
	}
	return $result
}

# =========================================================================

proc spec_get_super {} {
	global specstate
	
	set result {}
	set stateid [get_stateid spec]
	switch -- [lindex $stateid 0] \
		stateid {
			set tuples [lindex $stateid 1]
			foreach t $tuples {
				set info [lindex [lindex $stateid 1] 0]
# puts stderr "spec_get_super:all (lindex $stateid 1)"
# puts stderr "spec_get_super: $info ([lindex $info 1])"
				set supidx [lsearch $t super]
				set supnr [lindex $t [expr $supidx + 1]]
				if {$supidx == 0} {
					lappend result $supnr
				} else {
					lappend result [concat $supnr [lindex $t 0]]
				}
			}
			set specstate(cur) [join $result " "]
			if {![info exists specstate($specstate(cur))]} {
				set specstate($specstate(cur)) 1
				if {[info exists specstate(fid)]} {
					catch {
						puts $specstate(fid) "$specstate(cur)\t$tuples"
						flush $specstate(fid)
					}
				}
			}
			return $specstate(cur)
		} error {
			return $result
		}
}

proc spec_set_super {S} {
	global specstate
	
	if {[string compare $S $specstate(cur)] == 0} {
		puts stderr "skipping nop spec_set_super $S"
	} else {
		puts stderr "doing spec_set_super $S"
		set stateid [set_stateid spec $S]
		switch -- [lindex $stateid 0] \
			goto {
				set specstate(cur) $S
				return ""
			} goto_error {
				fatal "error from specification\n"
			}
	}
}

proc log_time_and_state {} {
	global maininfo
	utlog_tofile_timestamp
	set stats [spec_get_stats]
	switch -- [lindex $stats 0] \
		stats {
			utlog_tofile_stats [lindex $stats 1]
		} error {
			fatal "error from specification\n"
		}
	if {![info exists maininfo(nostateid,spec)]} {
		set stateid [get_stateid spec]
		switch -- [lindex $stateid 0] \
			stateid {
				utlog_tofile_stateid [lindex $stateid 1] SPEC
			} error {
				set maininfo(nostateid,spec) 1
			}
	}
	if {![info exists maininfo(nostateid,adapter)]} {
		set stateid [get_stateid adaptor]
		switch -- [lindex $stateid 0] \
			stateid {
				utlog_tofile_stateid [lindex $stateid 1] ADAPTOR
			} error {
				set maininfo(nostateid,adapter) 1
			}
	}
}

# =========================================================================

# Random number generator taken from:
#
# Brent Welch, "Practical Programming in Tcl/Tk", 2nd edition, pages 78,79
#   (adapted from Don Libes, "Exploring Expect", 1995 and from
#                 Press et al. "Numerical Recipes in C", 1998
#   )
#
proc RandomInit {seed} {
	global randomSeed
	set randomSeed $seed
}

proc Random {} {
	global randomSeed
	set randomSeed [expr ($randomSeed*9301 + 49297) % 233280]
	return [expr $randomSeed/double(233280)]
}

proc RandomRange {range} {
	return [expr int ([Random]*$range)]
}

# =========================================================================
proc default_channel { side } {
	switch -- $side {
		  input {
			set channel in
		} output {
			set channel out
		} default {
			set channel ""
		}
	}
}

# =========================================================================
