# =========================================================================
#
# simple command interpreter for manual I/O
#
# In a first version, we will only handle input.

# The idea is that this will allow us to ask the user to actually
# 'execute' a given stimulus, and after executing the stimulus confirm 
# execution of the stimulus to torx.
# This code will be called from the Adapter, via a specific user-interface
# channel. Probably, we may have several distinguishable PCO's that
# all 'use' a user-interface channel.
#
# Basically, for input we need one function, that first prints/shows
# a message to the user, telling the user what to do, and how to confirm
# that the stimulus is given, and then waiting until the user has
# given the confirmation, after which we return (with a status?).
#

proc ui_manual_stimulus { pco message {finalization ""} } {
	global maininfo uiinfo
	set uiinfo(timeout) -1 ;# disable timeout
	if {![info exists uiinfo(confirm)]} {
		# set uiinfo(confirm) uiok
		set uiinfo(confirm) ""
	}

	send_user "\ntorx_ui_manual_stimulus_start\n\n"
	send_user "$message\n"
	send_user "torx_ui_manual_stimulus_prompt $uiinfo(confirm)\n\n"
	send_user "When the stimulus has been given,\n"
        if {[string match $uiinfo(confirm) ""]} {
		send_user "type <return>."
	} else {
		send_user "type the following word, followed by <return>:"
		send_user " $uiinfo(confirm)"
	}
	send_user "\n\n"
	set go_on 1
	set result 0
	while {$go_on} {
		if {[string match $uiinfo(confirm) ""]} {
			send_user "type <return>: "
		} else {
			send_user "type $uiinfo(confirm): "
		}
		set timeout $uiinfo(timeout)
		# exp_internal 1
		set confirm ""
		expect_user {
			-re "^(\[^\r\n]*)\r*\n" {
				set confirm [string trim $expect_out(1,string)]
				if {[string match $confirm $uiinfo(confirm)]} {
					set go_on 0
					set result 1
				}
			}
			timeout {
			}
			eof {
				set maininfo(stopnow) 1
				send_user "\n"
				set go_on 0
			}
		}
		# exp_internal 0
		set timeout $maininfo(timeout)
	}
	if {![string match $finalization ""]} {
	  send_user "\ntorx_ui_manual_stimulus_finalization_start\n\n"
	  send_user "To finish giving the stimilus:\n"
	  send_user "$finalization\n"
	  send_user "torx_ui_manual_stimulus_finalization_end"
	}
	send_user "\ntorx_ui_manual_stimulus_end\n"
	return $result
}
