#! @PERL@
#! /usr/bin/perl

# This is like torx-logclient.expect --
# only as it seems, on cygwin xp this perl
# is much more reliable than tclsh or expect.

use Socket;

#! @PERL@

$debug = 0;
$BUFSIZ = 1024;				# size of chunks we try to read

$outpfx = '';
$errpfx = 'err:';

$| = 1; # do not buffer!

sub usage {
       print STDERR "usage: torx-logclient [ -m mcastid ] host port -- command args ...\n"
}

# print "ARGcnt ".int(@ARGV)." 0=".$ARGV[0]." 1=".$ARGV[1]."\n";

if (@ARGV <4) {
	usage();
	exit
}

if ($ARGV[0] eq "-m") {
	shift(@ARGV);
	$ENV{TORXMCASTID} = shift(@ARGV);
}

if (@ARGV <4) {
	usage();
	exit
}

$host = shift(@ARGV);
if ($host eq "0.0.0.0") {
	$host = "127.0.0.1";
}

if ($host eq "127.0.0.1" && defined($ENV{TORXMCASTID})) {
	$mcast = $ENV{TORXMCASTID};
	$mcast =~ s/!0\.0\.0\.0!/!127.0.0.1!/;
	$ENV{TORXMCASTID} = $mcast;
}
$port = shift(@ARGV);

if ($ARGV[0] ne "--") {
	usage();
	exit
}
shift(@ARGV);

$prog=shift(@ARGV);
$progname = $prog;
$progname =~ s,.*/,,;


dbgmsg("host=$host port=$port TORXMCASTID=$ENV{TORXMCASTID} pnm=$progname prog=$prog args=@ARGV");
start($progname, $prog, @ARGV);

$iaddr = inet_aton($host);
$paddr = sockaddr_in($port, $iaddr);

$proto = getprotobyname('tcp');
socket(SOCK, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
connect(SOCK, $paddr)   or die "connect: $!";


	$old = select(SOCK);
	$| = 1;
	select($old);

main_loop();

$SIG{CHLD} = sub { exit };

sub dbgmsg {
	my($s) = @_;
	print STDERR $$." ".$s."\n" if $debug;
}

sub repr {
	my($s) = @_;

	$s =~ s/\n/\\n/g;
	return $s;
}

sub main_loop {
	my($strReturns) = @_;
	my($ev, $time, $suspension);
	my($buffer,$line,$rest,$nread,$nfound,$stdinbuf,$iutoutbuf,$iuterrbuf);
	my($rin,$win,$ein,$rout,$wout,$eout);
	my(@words,$word);

	# $eofseen = "";

	# dbgmsg("STDIN main_loop");
	$timeout = undef;

	while(1) {
		$rin = $win = $ein = "";
		vec($rin, fileno(SOCK), 1 ) = 1 unless $eofseen{SOCK};
		vec($win, fileno(SOCK), 1) = 1;
		vec($rin, fileno(IUT_OUT), 1) = 1 unless $eofseen{IUT_OUT};
		vec($rin, fileno(IUT_ERR), 1) = 1 unless $eofseen{IUT_ERR};
		$ein = $rin; # | $win;
		if ($nfound = select($rout=$rin, undef, $eout=$ein, $timeout) < 0) {
			die "select: $!";
		} else {
			# dbgmsg("STDIN main_loop after select");
		}
		if (vec($rout, fileno(SOCK), 1)) {
			# dbgmsg("SOCK ready");
			$buffer = '';
			$stdinbuf .= doread(SOCK);
			dbgmsg("SOCK stdinbuf (".repr($stdinbuf).")");
			$stdinbuf = dowrite(IUT_IN, $stdinbuf, '', $eofseen{SOCK});
			if ($eofseen{SOCK}) {
				close(IUT_IN);
				$timeout = 1;
				exit;
			}
			# dbgmsg("SOCK handled");
		}
		if (vec($rout, fileno(IUT_OUT), 1)) {
			# dbgmsg("IUT_OUT ready");
			$iutoutbuf .= doread(IUT_OUT);
			dbgmsg("IUT_OUT iutoutbuf (".repr($iutoutbuf).")");
			$iutoutbuf = dowrite(SOCK, $iutoutbuf, $outpfx, $eofseen{IUT_OUT});
			# dbgmsg("IUT_OUT handled");
		}
		if (vec($rout, fileno(IUT_ERR), 1)) {
			# dbgmsg("IUT_ERR ready main_loop");
			$iuterrbuf .= doread(IUT_ERR);
			dbgmsg("IUT_ERR iuterrbuf (".repr($iuterrbuf).")");
			$iuterrbuf = dowrite(SOCK, $iuterrbuf, $errpfx, $eofseen{IUT_ERR});
			# dbgmsg("IUT_ERR handled");
		}
		handle_eof_iut();
		if ($nfound == 0 && defined($timeout)) {
			cleanup;
			exit;
		}
	}
}

sub doread {
	my($fd) = @_;

	my($buffer, $nread) = '';
	if (($nread = sysread($fd, $buffer, $BUFSIZ)) > 0) {
		# dbgmsg("$fd read $nread");
	} elsif ($nread == 0) {
		dbgmsg("$fd eof");
		$eofseen{$fd} = 1;
	}
	return $buffer;
}

sub dowrite {
	my($fd, $buf, $pfx, $all) = @_;
	my($line,$rest);

	while ($buf =~ /^(([^\n\r]*)\r*\n)/) {
		($line,$rest) = ($2, $');
		# $line =~ s/^[ 	]*//;
		# $line =~ s/[ 	]*$//;
		print $fd $pfx.$line."\n";
		$buf = $rest;
	}
	# if ($buf ne "" && $all) {
# 		$buf = dowrite($fd, $buf."\n", $pfx, 0);
# 	}
	return $buf;
}

sub handle_eof_iut {
	# AFEB hack: changed && to || to semi-detect eof of torx, also on windows
	if ($eofseen{IUT_OUT} || $eofseen{IUT_ERR}) {
		cleanup;
		exit;
	}
}

sub cleanup {
}


sub start {
	my($cmd, @cmd) = @_;
	my($old, $sh, @shcmd);
	
	dbgmsg("start $cmd : ".join(" ", @cmd));
	$sh = "/bin/sh";
	@shcmd = ($sh, "-c", join(" ", @cmd));
	dbgmsg("shstart $sh : ".join(" ", @shcmd));
	pipe(IUT_OUT,WOFD);    	# Open pipes that will link child and parent
	pipe(IUT_ERR,WEFD);
	pipe(WIFD,IUT_IN);
	if (($cpid = fork()) == 0) {	#in child
		close(IUT_IN);
		close(IUT_OUT);
		close(IUT_ERR);
		close(STDOUT);
		open(STDOUT,">&WOFD");
		$| = 1;						# is this really needed?
		close(STDERR);
		open(STDERR,">&WEFD");
		close(STDIN);
		open(STDIN,"<&WIFD");
		$| = 1;						# is this really needed?
		exec $sh @shcmd;				# do it
		die "exec: $cmd: $!\n";		# or die
	}
	# in parent
	close(WIFD);
	close(WOFD);
	close(WEFD);
    
	# seems we need the following to enable autoflush
	$old = select(IUT_IN);
	$| = 1;
	select($old);
	dbgmsg("started $cmd : ".join(" ", @cmd));
}

# some support routines for the select code.
sub addBits {
    local($fd) = @_;
    vec($fileDescs, $fd, 1) = 1;
    $fds{$fd} = 1;
}
sub delBits {
    local($fd) = @_;
    vec($fileDescs, $fd, 1) = 0;
    delete $fds{$fd};
}
sub allEofBits {
    ! keys %fds;
}
sub printVec {
    local($v) = @_;
    local($i, $result);

    for ($i = (8*length($v)) - 1; $i >= 0; $i--) {
	$result .= (vec($v, $i, 1)) ? "1" : "0";
    }
    $result;
}

