#! @PERL@ --

#
# usage: udp [ -port portnr ] [ -[no]printdata ] [ -[no]printdatahex ] [ -[no]delay ]
# 
# DECRIPTION:
#
# opens given port, or the first free one.
# prints on stdout the number of the opened port, followed by newline
#
# Now, wait for something to happen on stdin and on port.
#  - expect on stdin a line of text like:
#         hostname portnr message newline
#    where hostname is hostname or FQDN or decimal-dot notation
#          the message is encoded in hex
#    do send the message to the hostname:port
#  - when something happens on port:
#    receive packet, and output on stdout a line of the form:
#         decimal-dot-host portnr message newline
#    where decimal-dot-host is sending host in decimal-dot notation
#          the message is encoded in hex
# Do this until we are killed
# (or EOF on stdin, or a special quit command is given?)
#
#
# use Socket;
# 
# socket(SERVER, PF_INET, SOCK_STREAM, getprotobyname('tcp'));
# setsockopt(SERVER, SOL_SOCKET, SO_REUSEADDR, 1);
# bind(SERVER, sockaddr_in($server_port, INADDR_ANY))
#     or die "Binding: $!\n";
# 
# # accept loop
# while (accept(CLIENT, SERVER)) {
#     $my_socket_address = getsockname(CLIENT);
#     ($port, $myaddr)   = sockaddr_in($my_socket_address);
# }


use strict;

require 5.002;
use Socket;
use Sys::Hostname;
use Getopt::Long;

my ( $prog, $max_udp_len, $proto, $BUFSIZ,
     $peer_iaddr, $peer_paddr, $peer_host, $peer_port, $peer_addr,
     $iaddr, $paddr, $host, $port, $addr,
     $rin, $rout, $win, $wout, $ein, $eout,
     %optcl, $result, $debug, $printdata, $printdatahex,
     $command, $buffer, $bufferhex, %buf, $nfound, $nread,
     $message_hex, $message, $delay, $delayprob, $delaytime
   );

$prog = 'udp';	# should be something like `basename $0`
$port = 0;		# we don't care, or the port given on the command line
$BUFSIZ = 1024;
$max_udp_len = 1500;

# allow random delay (sleep) between receive on stdin and send on socket
# to simulate delaying (FIFO buffering) channels
$delay = 0;		# boolean, should we delay at all?
$delayprob = 0.5;	# we delay only if random() > $delayprob
$delaytime = 1;	# time to sleep if we delay

%optcl = ();
$result = GetOptions(
			\%optcl,
			"port=i",
			"debug:i",
			"delay!",
			"printdata!",
			"printdatahex!"
		);

die "usage: $0 [ -port portnr ] [ -[no]printdata ] [ -[no]printdatahex ] [ -[no]delay ]\n"
	unless $result;

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});
$port = $optcl{"port"} if ((defined $optcl{"port"}) && $optcl{"port"});
$printdata = $optcl{"printdata"};
$printdatahex= 1 unless ((defined $optcl{"printdatahex"}) && ! $optcl{"printdatahex"});
$delay = $optcl{"delay"};

if ($debug) {
	print $prog.": ($result)(".$optcl{"port"}.")(".$optcl{"debug"}.")(".$optcl{"printdata"}.")(".$optcl{"printdatahex"}.")\n";
	print $prog.": ($result)(".$port.")(".$debug.")(".$printdata.")(".$printdatahex.")(".$delay.")\n";
}
exit if $debug >= 3;

# $iaddr = gethostbyname("localhost");
# $iaddr = pack("C4", (0, 0, 0, 0));
$iaddr = INADDR_ANY;
$proto = getprotobyname('udp');
$paddr = sockaddr_in($port, $iaddr); # if $port = 0 we don't care

socket(SOCKET, PF_INET, SOCK_DGRAM, $proto)		or die "$prog: socket: $!";
bind(SOCKET, $paddr)							or die "$prog: bind: $!";

$paddr = getsockname(SOCKET);			# get chosen port
($port, $iaddr) = sockaddr_in($paddr);	# get chosen port
$host = gethostbyaddr($iaddr, AF_INET);
$addr = inet_ntoa($iaddr);

# setsockopt(SOCKET, SOL_SOCKET, SO_REUSEADDR, 1)
# 	or warn "$prog: setsockopt: $!\n";

$| = 1;				# don't buffer output on stdout
print "HOSTPORT ".$host." ".$addr." ".$port."\n";	# let the world know on which port we are

$rin = $win = $ein = "";
$rin = fdbits(fileno(STDIN), fileno(SOCKET));
$ein = $rin | $win;

$buf{'STDIN'} = "";

my $oldfh;
$oldfh = select STDIN; $| = 1; select $oldfh;
$oldfh = select STDOUT; $| = 1; select $oldfh;
$oldfh = select STDERR; $| = 1; select $oldfh;
$oldfh = select SOCKET; $| = 1; select $oldfh;

$debug = 0;
if ($debug) {
	open(DEBUG, ">/dev/pts/10");
	$oldfh = select DEBUG; $| = 1; select $oldfh;
}


for (;;) {
	print DEBUG "$port fileDescs:  ", &printVec($rin), "\n"
	    if $debug;
	if (($nfound = select($rout=$rin, $wout=$win, $eout=$ein, undef)) < 0) {
	    die "$prog: $port: select: $!";
	}

	print DEBUG "rout:  ", &printVec($rout), ", " if $debug;
	print DEBUG "$port: rout:  ", &printVec($rout), "\n" if $debug;
	print DEBUG "$port: wout:  ", &printVec($wout), "\n" if $debug;
	print DEBUG "$port: eout:  ", &printVec($eout), "\n" if $debug;
	
	if (vec($eout, fileno(main::STDIN), 1)) {
		    print DEBUG "$port: STDIN eof\n" if $debug;
		    close(SOCKET);
		    exit(0);
	}
	if (vec($eout, fileno(SOCKET), 1)) {
		    print DEBUG "$port: SOCK eof\n" if $debug;
		    close(SOCKET);
		    exit(0);
	}
	if (vec($rout, fileno(main::STDIN), 1)) {
		print DEBUG "$port: STDIN ready\n" if $debug;
		$buffer = '';
		if (($nread = sysread(main::STDIN, $buffer, $BUFSIZ)) > 0) {
		    $buf{'STDIN'} .= $buffer;
		    print DEBUG "$port: STDIN read $nread\n" if $debug;
		} elsif ($nread == 0) {
		    print DEBUG "$port: STDIN eof\n" if $debug;
		    close(SOCKET);
		    exit(0);
		}
		if ($buf{'STDIN'} =~ /^([^\n]*)\n/) {
			$command = $1;
			$buf{'STDIN'} = $';
			if ($command =~ /^SENDHEX[ \t]+([a-zA-Z0-9._]+)[ \t]+([0-9]+)([ \t]+([0-9a-fA-F]+))?$/i) {
				$peer_host = $1;
				$peer_port = $2;
				$message_hex = $4;
			
				if (! ($peer_iaddr = inet_aton($peer_host))) {
					warn "$prog: unknown host $peer_host\n";
					next;
				}
				$peer_paddr = sockaddr_in($peer_port, $peer_iaddr);
				$message = pack("H*", $message_hex);
				if ($delay && rand() > $delayprob) {
					print STDERR "DEBUG $prog: sleep $delaytime before sending to $peer_host:$peer_port\n";
					sleep($delaytime);
				}
				if (!defined(send(SOCKET, $message, 0, $peer_paddr))) {
					warn "$prog: send $peer_host:$peer_port: $!\n";
				} else {
					print DEBUG "SOCKET send ".length($message)."\n" if $debug;
				}
			} elsif ($command =~ /^LOCALADDR$/i) {
				print "HOSTPORT ".$host." ".$addr." ".$port."\n";
			} elsif ($command =~ /^PRINTDATA$/i) {
				$printdata = 1;
			} elsif ($command =~ /^NOPRINTDATA$/i) {
				$printdata = 0;
			} elsif ($command =~ /^PRINTDATAHEX$/i) {
				$printdatahex = 1;
			} elsif ($command =~ /^NOPRINTDATAHEX$/i) {
				$printdatahex = 0;
			} elsif ($command =~ /^DEBUG([ \t]+([0-9]+))?$/i) {
				if ($2 ne "") {
					$debug = $2;
				} else {
					$debug = 1;
				}
			} elsif ($command =~ /^NODEBUG$/i) {
				$debug = 0;
			}
		}
	}
	if (vec($rout, fileno(SOCKET), 1)) {
		print DEBUG "$port: SOCKET ready\n" if $debug;
		$buffer = '';
		($peer_paddr = recv(SOCKET, $buffer, $max_udp_len, 0))
			or warn "$prog: recv: $!";
		print DEBUG "SOCKET read ".length($buffer)."\n" if $debug;
		($peer_port, $peer_iaddr) = sockaddr_in($peer_paddr);
		$peer_host = gethostbyaddr($peer_iaddr, AF_INET);
		$peer_addr = inet_ntoa($peer_iaddr);
		# HACK for localhost just in case reverse lookup fails
		# which might happen in incomplete envs (need seen on cygwin)
		if ($peer_host eq '' && $peer_addr eq '127.0.0.1') {
			$peer_host = 'localhost';
		}
		$bufferhex = unpack("H*", $buffer);
		print STDOUT "RECVHEX ".$peer_host.' '.$peer_addr.' '.$peer_port.' '.$bufferhex."\n"
			if $printdatahex;
		print STDOUT "RECV ".$peer_host.' '.$peer_addr.' '.$peer_port.' '.$buffer."\n"
			if $printdata;
	}
}


# some support routines for the select code.
sub printVec {
    my($v) = @_;
    my($i, $result);

    for ($i = (8*length($v)) - 1; $i >= 0; $i--) {
	$result .= (vec($v, $i, 1)) ? "1" : "0";
    }
    $result;
}

sub fhbits {
	my(@fhlist) = @_;
	my($bits);
	for (@fhlist) {
		vec($bits, fileno($_), 1) = 1;
	}
	return $bits;
}
sub fdbits {
	my(@fdlist) = @_;
	my($bits);
	for (@fdlist) {
		vec($bits, $_, 1) = 1;
	}
	return $bits;
}
