#
# expcadp.pm - mkprimer module for the CADP LOTOS explorer
# will be ``required'' by the mkprimer program
#
package expcadp;

$libname = 'libexpcadp';
$gcname = 'gc';
$expsuffix = '.tx';


if ( -r "$libdir/lib".$gcname.".a") {
  $gc_flags = "-L$libdir -l$gcname";
} else {
  $gc_flags = "";
}

%suffix = (
	     );

%lang = (
	"LOTOS-EXPLORER" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix, $paramsref) = @_;
		my(%params) = %$paramsref;
		my($flag) = "-f $params{config}" if $params{config} ne "";
		my(@cmds) = (
			"caesar.open $specfile $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$expsuffix$exesuffix",
			"echo '#!/bin/sh' > $target",
			"echo exec primer $flag \\\"\\\$@\\\" $target$expsuffix$exesuffix >> $target",
			"chmod 755 $target",
		          );
		::Exec(@cmds);
	},
	"BCG-EXPLORER" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_open $specfile $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$expsuffix$exesuffix",
			"echo '#!/bin/sh' > $target",
			"echo exec primer $flag \\\"\\\$@\\\" $target$expsuffix$exesuffix >> $target",
			"chmod 755 $target",
		          );
		::Exec(@cmds);
	},
	"AUT-CADP-EXPLORER" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_io -aldebaran - -bcg $specpath$specname.bcg < $specfile",
			"bcg_open $specpath$specname.bcg $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$expsuffix$exesuffix",
			"echo '#!/bin/sh' > $target",
			"echo exec primer $flag \\\"\\\$@\\\" $target$expsuffix$exesuffix >> $target",
			"chmod 755 $target",
		          );
		::Exec(@cmds);
	},
	"FC2-EXPLORER" =>
	 sub  {
		my($specfile, $target, $prefix, $libdir,
		   $specpath, $specname, $specsuffix, $exesuffix) = @_;
		my(@cmds) = (
			"bcg_io -fc2 - -bcg $specpath$specname.bcg < $specfile",
			"bcg_open $specpath$specname.bcg $gc_flags $libdir/$libname.a < /dev/null",
			"mv $libname$exesuffix $target$expsuffix$exesuffix",
			"echo '#!/bin/sh' > $target",
			"echo exec primer $flag \\\"\\\$@\\\" $target$expsuffix$exesuffix >> $target",
			"chmod 755 $target",
		          );
		::Exec(@cmds);
	},
);

::Register(\%lang, \%suffix);

1;
