# =========================================================================
# $Id: main_loop.tcl,v 1.1.2.6 2008/07/07 02:30:17 belinfan Exp $
# =========================================================================

proc explorer_main {} {
	global cont
	if {[catch {
		## Currently, we prefere to use blocking IO
		## fconfigure -nonblocking stdin
		fileevent stdin readable explorer_try_handle_line
		vwait cont
	} foo]} {
		bgerror $foo
	}
}

# ====================================

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc bgerror {msg} {
	global errorInfo
	debugMsg "Fatal InternalError: $msg"
	set lines [split $errorInfo \n]
	debugMsg "Stack Begin"
	foreach l $lines {
		debugMsg "Stack: $l"
	}
	debugMsg "Stack End"
	# exit
}


# ====================================

proc explorer_try_handle_line {} {
	fileevent stdin readable ""
	set stop 0
	set skip 0
	set line ""
	if {[gets stdin line] < 0} {
		if {[fblocked stdin]} {
			set skip 1
		} elseif {[eof stdin]} {
			set line "q"
			set stop 1
		}
		# HACK
		exit
	} 
	if {! $skip} {
		set speed [time {explorer_handle_line $line}]
		puts "A_DEBUG time: $speed"
	}
	if {! $stop} {
		fileevent stdin readable explorer_try_handle_line
	}
}

# ====================================

proc explorer_handle_line {line} {
	global cont
	# debugMsg "Line: ($line)"
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [parse_line $line]
		# debugMsg "Parsed: ($parsed_line)"
		set cmd_keyword [lindex $parsed_line 0]
		switch -- $cmd_keyword {
		    q {
		    	set cont 0
			   	puts "Q"
		  } r {
		  		set initial [explorer_start]
		  		puts "R [join $initial \t]"
		  } e {
		  		set s [lindex $parsed_line 1]
		  		
		  		if {[catch {explorer_unfold $s} evs]} {
			  		puts "E0 $evs"
		  		} else {
			  		puts "EB"
			  		foreach e $evs {
			  			puts "Ee\t[join $e \t]"
			  		}
			  		puts "EE"
			  	}
		  } d {
		  		# nothing to do here; in  general: delete list of events/states
		  		puts "D"
		  } default {
				puts "A_ERROR UnknownCommand: $cmd_keyword"
		  }
		}
	}
}

# =========================================================================

proc parse_line {line} {
	if {[regexp {^(.)([ \t]+(.*))?$} $line match cmd dummy remains]} {
		return [list $cmd $remains]
	}
}

proc split_in_two_on_ws {line} {
	if {[regexp {^([^ \t]+)([ \t]+(.*))?$} $line match cmd dummy remains]} {
		return [list $cmd $remains]
	}
}

# =========================================================================
