#include "k.h"  /* includes other include files, for strings etc */
#include "buf.h"

static void bufreserve(Buf*, int);
static void bufgrow(Buf*, int);
static void bufdocopy(Buf*, char*, int);

Buf*
bufnew(void)
{
	Buf *b;

	b = emalloc(sizeof(Buf));
	b->b = emalloc(1);
	b->sz = 1;
	*b->b = '\0';
	b->e = b->b;
	return b;
}

char*
bufstring(Buf *b)
{
	if (b == 0)
		return 0; /* should report error */

	return b->b;
}

void
bufaddc(Buf *b, char c)
{
	if (b == 0)
		return; /* should report error */

	bufdocopy(b, &c, 1);
}

void
bufcat(Buf *b, char *s)
{
	int l;

	if (b == 0)
		return; /* should report error */

	l = strlen(s);
	bufdocopy(b, s, l);
}

void
bufcpy(Buf *b, char *s)
{
	int l;

	if (b == 0)
		return; /* should report error */

	b->e = b->b;
	l = strlen(s);
	bufdocopy(b, s, l);
}

static void
bufreserve(Buf *b, int n)
{
	if (b == 0)
		return; /* should report error */

	bufgrow(b, (b->e - b->b) + n + 1); /* +1 for trailing \0 */
}

static void
bufgrow(Buf *b, int n)
{
	int l;

	if (b == 0)
		return; /* should report error */

	if (n > b->sz) {
		l = b->e - b->b;
		b->b = erealloc(b->b, n);
		b->e = b->b + l;
		b->sz = n;
	}
}

static void
bufdocopy(Buf *b, char *s, int l)
{
	if (b == 0)
		return; /* should report error */

	bufreserve(b, l);
	memcpy(b->e, s, l);
	b->e += l;
	*b->e = '\0';
}
