#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

void init_mtab(void);
int mfind(const char *);
char *mstring(int);

void printm(int x);

char **mtab = NULL; // Table for storing mtypes. Instantiated by init_mtab().
int mtab_size = 0;  // Size of mtab table
int mtab_index = 0; // Current index in mtab table.


void init_mtab(void) {
    int    saved_stdout_fd;
    fpos_t pos;
        
    int pipe_fd[2];
    
    char buf[80];
    
    int i; // amount of characters read()
    
    char **tmp;


    fflush(stdout);
    fgetpos(stdout, &pos);			// Save position
    saved_stdout_fd = dup(fileno(stdout));	// Duplicate stdout filedescriptor (to save it)
        
    pipe(pipe_fd); // Create new pipe. read on pipe_fd[0], write on pipe_fd[1]
    
    dup2(pipe_fd[1], fileno(stdout)); // Replace stdout(fd) with write-pipe fd
    
    if (mtab == NULL) {
    	mtab_size = 1;
    	mtab = calloc(mtab_size, sizeof(char *));
    	if ( mtab == NULL) {
    		fprintf(stderr, "Error. Could not allocate memory for mtype table. Aborting.\n");
    		exit(-1);
    	}
    	// 0 is not used as an index for a mtype
    	mtab[0] = "";
    }
    	
    
    
    for ( mtab_index = 1; ; mtab_index++) {
    	// Ask pan to print (to stdout!) mtype 'i'
    	printm(mtab_index);
    
    	fflush(stdout);

    	i = read(pipe_fd[0], buf, 79);
    	buf[i] = '\0';
    	
    	// If no label for 'i' exists, pan will just print the number.
    	// If this happens, it means we're done with filling the mtab.
    	if (atoi(buf) == mtab_index ) {
    		break;
    	}
    	
    	if (mtab_index >= mtab_size) {
    		mtab_size = mtab_size + 2;
    		tmp = realloc(mtab, mtab_size * sizeof(char *));
    		if (tmp == NULL) {
    			fprintf(stderr, "Error. Could not allocate memory to increase mtype table. Aborting.\n");
    			exit(-1);
    		}
    		mtab = tmp;
    	}
    	mtab[mtab_index] = strdup(buf);
    	
    }
    // end for
    
    close(pipe_fd[0]);
    close(pipe_fd[1]);

    dup2(saved_stdout_fd, fileno(stdout));	// Replace write-pipe fd with saved stdout fd
        
    close(saved_stdout_fd); // Close the copy of stdout
        
    clearerr(stdout);
    fsetpos(stdout, &pos);	
	
}

int mfind(const char *mtype) {
	int i;
	
	for ( i = 1; i < mtab_index; i++ ) {
		if (strcmp(mtype, mtab[i]) == 0) {
			return i;
		}
	}
	return -1;
}

char *mstring(int index) {
	if (index >= 0 && index < mtab_index) {
		return mtab[index];
	} else {
		return "";
	}
}
