# =========================================================================
#
# $Id: primer.tcl,v 1.2.2.5 2006/01/24 21:42:37 belinfan Exp $
#
# Author: Axel Belinfante
#
# This file is _VERY_ simular to the mutant.tcl file!
# If you change something here, changes are VERY high that you have to
# change it in mutant.tcl as well!
#
# This file implements the PRIMER command that we use in primer
# config files.
# The idea is to interpret primer files as tcl files containing
# only invocations of the PRIMER command, to register primers.
# As result of such registration, the GUI will add the registered
# items to the primers menu.
# Question: will the loading of a second primer file override the first one?
# Or can we have items of multiple primer files 'active' at the same time?
# If so, how will this be visible in the menu?
# (as a sub-menu under the Primers menu?)
# How will we be able to 'close' different primer files?
# (via a sub-menu under the Close Primers menu entry?)
#
# =========================================================================

# This routine will be invoked from xtorx to force auto-loading of this file.
proc do_first_initPrimers {args} {
	do_first_initExtension $args
}

proc do_initPrimers {args} {
	global tester
	global debug

	# set debug(Msg,do_extension_callback) 1

	do_initExtension \
		primers \
		PRIMER \
		"*primers" \
		"Open Primers File" \
		tester(primers,titlename) \
		tester(basenamebare,primers_file) \
		tester(torx,cmdline,primers,flags) \
		tester(torx,cmdline,primers,config) \
		"primers_present *primers_mb,*primers,*close_primers_button,*reopen_primers_button" \
		"primers_present *open_primers_button" \
		{
			{{Primers Files}    {.primers}    }
			{{Guides Files}    {.guides}    }
			{{Mutants Files}    {.mutants}    }
			{{Trace Files}      {.trace}      }
			{{Log Files}        {.log}        }
			{{Config Files}     {.if}         }
			{{Text Files}       {.txt}        }
			{{TCL Scripts}      {.tcl}        }
			{{All Files}        *             }
		} \
		$args
}

proc do_cleanupPrimers {args} {
	do_cleanupExtension primers $args
}

proc do_open_primers {args} {
	do_open_extension primers $args
}

proc do_do_open_primers { filename} {
	do_do_open_extension primers $filename
}

proc do_close_primers {args} {
	do_close_extension primers $args
}

proc do_do_close_primers {} {
	do_do_close_extension primers
}

proc do_reopen_primers {args} {
	do_reopen_extension primers $args
}

proc do_do_select_primer {titlename} {
	do_do_select_extension primers $titlename
}

proc PRIMER {entry title create destruct args} {
	EXTENSION primers $entry $title $create $destruct $args
}
