proc update_menu {} {
	global maininfo specinfo menuinfo
	event_table_clear
	foreach p $specinfo(primerids) {
		# incredible hack
		# to get this primer added to the list of primers we intersect on
		event_table_register inputs $p
		set result [spec_get_inputs $p]
		if {$maininfo(debug) > 2} {
			send_error "DEBUG: spec_get_inputs result ($p): $result\n"
		}
		switch -- [lindex $result 0] \
			inputs {
				set events [lindex $result 1]
				foreach ep $events {
					set e [lindex $ep 0]
					set s [lindex $ep 1]
					set a [lindex $ep 2]
					event_table_insert inputs $p $e $a
				}
			} error {
			}
	}
	
	# while handling outputs, we have to take out the events
	# that represent something special of primers in roles test and guide,
	# because these have to be interpreted in a special way:
	#  -  $specinfo(label,hit,$p)   => signal succesfull end-of-trace
	#  -  $specinfo(label,miss,$p)  => signal unsuccesfull end-of-trace
	#  -  $specinfo(label,delta,$p) => to be mapped on Delta,
	#                                  i.e. on $specinfo(label,delta,this)
	#
	# In addition, we keep track of having seen a Verdict,
	# because if we see (and inform the user) of a verdict here,
	# we may want to suppress a verdict based on (e.g.) an empty
	# menu...
	set menuinfo(verdict) ""
	set menuinfo(verdict,seen) 0
	foreach p $specinfo(primerids) {
		# incredible hack
		# to get this primer added to the list of primers we intersect on
		event_table_register outputs $p
		set result [spec_get_outputs $p]
		if {$maininfo(debug) > 2} {
			send_error "DEBUG: spec_get_outputs result ($p): $result\n"
		}
		switch -- [lindex $result 0] \
			outputs {
				set events [lindex $result 1]
				if {   [string match $specinfo(role,$p) test]
				    || [string match $specinfo(role,$p) guide]} {
					foreach ep $events {
						set e [lindex $ep 0]
						set susp [lindex $ep 1]
						if {![is_special $p $e]} {
							if {[info exists specinfo(label,delta,$p)]} {
								if {[event_eq $specinfo(label,delta,$p) $e]} {
									if {$maininfo(debug) > 2} {
										send_error "DEBUG: delta match: ($specinfo(label,delta,$p)) eq ($e)\n"
									}
									if {[info exists specinfo(label,delta,this)]} {
										event_table_insert outputs $p $specinfo(label,delta,this) [list [list suspension 1]]
									} else {
										event_table_insert outputs $p Delta [list [list suspension 1]]
									}
									continue
								}
							}
							event_table_insert outputs $p $e [list [list suspension 0]]
						}
					}
				} else {
					foreach ep $events {
						set e [lindex $ep 0]
						set susp [lindex $ep 1]
						if {[string compare $susp ""] != 0} {
							event_table_insert outputs $p $e [list [list suspension $susp]]
						} else {
							event_table_insert outputs $p $e
						}
					}
				}
			} error {
			}
	}
	set res_inputs [event_table_intersect inputs]
	set res_outputs [event_table_intersect outputs]
	set menuinfo(inputs) [lindex $res_inputs 2]
	set menuinfo(outputs) [lindex $res_outputs 2]
	set menuinfo(ninputs) [lindex $res_inputs 1]
	set menuinfo(noutputs) [lindex $res_outputs 1]
	set menuinfo(inputnprimers) [lindex $res_inputs 0]
	set menuinfo(outputnprimers) [lindex $res_outputs 0]
	
	set menuinfo(inconsistent) 0
	set menuinfo(endoftest) 0
	if {([llength $menuinfo(inputs)] == 0) &&
		([llength $menuinfo(outputs)] == 0)} {
		if {(($menuinfo(inputnprimers) > 1) &&
		     ($menuinfo(ninputs) > 0)) ||
		     (($menuinfo(outputnprimers) > 1) &&
		     ($menuinfo(noutputs) > 0))} {
			set menuinfo(inconsistent) 1
		} else {
			set menuinfo(endoftest) 1
		}
	}
	send_menu_verdict
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: menu...inputs:\n"
		foreach ev $menuinfo(inputs) {
			send_error "DEBUG: inputmenu: $ev\n"
		}
		send_error "DEBUG: menu...end inputs:\n"
		send_error "DEBUG: menu...outputs:\n"
		foreach ev $menuinfo(outputs) {
			send_error "DEBUG: inputmenu: $ev\n"
		}
		send_error "DEBUG: menu...end outputs:\n"
	}
}

proc is_special {p e} {
	global specinfo
	global menuinfo
	if {[info exists specinfo(label,hit,$p)]} {
		if {[event_eq $specinfo(label,hit,$p) $e]} {
			# end of trace; we will stop
			#send_user "A_GUIDANCE_VERDICT HIT\n"
			if {! $menuinfo(verdict,seen)} {
				set menuinfo(verdict) HIT
				set menuinfo(verdict,seen) 1
			}
			return 1
		}
	}
	if {[info exists specinfo(label,pass,$p)]} {
		if {[event_eq $specinfo(label,pass,$p) $e]} {
			# end of trace; we will stop
			#send_user "A_GUIDANCE_VERDICT PASS\n"
			if {! $menuinfo(verdict,seen)} {
				set menuinfo(verdict) PASS
				set menuinfo(verdict,seen) 1
			}
			return 1
		}
	}
	if {[info exists specinfo(label,miss,$p)]} {
		if {[event_eq $specinfo(label,miss,$p) $e]} {
			# end of trace; we will stop
			#send_user "A_GUIDANCE_VERDICT INCONCLUSIVE\n"
			# XXX What do we do here?
			set menuinfo(verdict) INCONCLUSIVE
			set menuinfo(verdict,seen) 1
			return 1
		}
	}
	# if we get here, we did not recognize the event
	# as 'special', so we simply add it to the output
	# menu
	return 0
}

proc menu_contains_hit {p} {
	global maininfo
	global specinfo
	set result [spec_get_outputs $p]
	set special_found 0
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: spec_get_outputs result ($p): $result\n"
	}
	switch -- [lindex $result 0] \
		outputs {
			set events [lindex $result 1]
			if {   [string match $specinfo(role,$p) test]
			    || [string match $specinfo(role,$p) guide]} {
				foreach ep $events {
					set e [lindex $ep 0]
					if {[is_special $p $e]} {
						set special_found 1
					}
				}
			} else {
				# internal error; we are expected to be called only for
				# guides!
			}
		} error {
		}
	return $special_found
}

proc event_in_menu { ev menu } {
	global maininfo
	set ev [event_canonical $ev]
	set result [lsearch -exact $menu $ev]
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: event_in_menu ($ev)($menu)([llength $menu]) result ($result)\n"
	}
	if {$result >= 0} {
		return 1
	}
	foreach el $menu {
		if {[regsub -all -- var_\[a-zA-Z\]+(\\\$\[0-9\]+)? $el * pat] > 0} {
			if {[string match $pat $ev]} {
				return 1
			}
		}
	}
	return 0
}

proc event_match { ev1 ev2 } {
	# NOTE: older versions of Tcl do not allow a '-nocase' flag
	# so we use 'string tolower' before comparing
	# (actually, we should be samrt and try to see if we
	#  can use the '-nocase' flag automatically )
	set ev1 [event_canonical $ev1]
	set ev2 [event_canonical $ev2]
	if {[string match $ev1 $ev2] || [string match $ev2 $ev1]} {
		return 1
	} elseif {[regsub -all -- var_\[a-zA-Z\]+(\\\$\[0-9\]+)? $ev1 * pat] > 0
			&& [string match $pat $ev2]} {
		return 1
	} elseif {[regsub -all -- var_\[a-zA-Z\]+(\\\$\[0-9\]+)? $ev2 * pat] > 0
			&& [string match $pat $ev1]} {
		return 1
	} else {
		return 0
	}
}

proc event_eq { ev1 ev2 } {
	# NOTE: we use 'string compare' instead of 'string match'
	# because 'string match pattern string' will treat certain
	# characters in 'pattern' as special, which is not what we
	# want here: we want a simple char-by-char string comparison.
	# NOTE: older versions of Tcl do not allow a '-nocase' flag
	# so we use 'string tolower' before comparing
	# (actually, we should be samrt and try to see if we
	#  can use the '-nocase' flag automatically )
	set ev1 [event_canonical $ev1]
	set ev2 [event_canonical $ev2]
	if {[string compare $ev1 $ev2] == 0} {
		return 1
	} else {
		return 0
	}
}

proc event_canonical { ev } {
	regsub -all -- { } $ev {} ev
	set ev [string tolower $ev]
	return $ev
}
