proc cache_init {{autname ""}} {
	global cache
	
	set cache(entries) 0
	set cache(hit) 0
	set cache(miss) 0
	set cache(insthit) 0
	set cache(instmiss) 0
	set cache(states) {}
	set cache(ntrans) 0
	set cache(start) 0
	set cache(faut) {}
	
	if {[string compare $autname ""] != 0} {
		if {[catch {open $autname w} fileId]} {
			puts stderr "cannot open $autname for writing: $fileId"
		} else {
			set cache(faut) $fileId
			puts $fileId "des(0, 0, 0)"
		}
	}
}

proc cache_explorer_unfold {s i si} {
	global cache
	
	set faut $cache(faut)
	if {[info exists cache(events,$s)]} {
		incr cache(hit)
		debugMsg "cache ($s=$i=$si) hit $cache(hit) $cache(miss) $cache(entries)"
		return $cache(events,$s)
	}
	if {[string compare $si ""] != 0} {
		if {[info exists cache(events,$si)]} {
			incr cache(hit)
			debugMsg "cache ($s=$i=$si) hit $cache(hit) $cache(miss) $cache(entries)"
			return $cache(events,$si)
		}
		lappend cache(states) $si
		set ev [explorer_unfold $si]
		cache_incr_dump $faut $si $ev
		set cache(events,$si) $ev
		incr cache(ntrans) [llength $ev]
		set cache(events,$s) $ev
		incr cache(entries)
		incr cache(miss)
		debugMsg "cache ($s=$i=$si) miss $cache(hit) $cache(miss) $cache(entries)"
		return $ev
	}
	lappend cache(states) $s
	set ev [explorer_unfold $s]
	cache_incr_dump $faut $s $ev
	set cache(events,$s) $ev
	incr cache(ntrans) [llength $ev]
	incr cache(entries)
	incr cache(miss)
	debugMsg "cache ($s=$i=$si) miss $cache(hit) $cache(miss) $cache(entries)"
	return $ev
}

proc cache_explorer_instantiate {s si preds} {
	global cache
	
	set faut $cache(faut)
	if {[info exists cache(inst,$s,$preds)]} {
		incr cache(insthit)
		debugMsg "cache ($s $preds) insthit $cache(insthit) $cache(instmiss) $cache(hit) $cache(miss) $cache(entries)"
		return $cache(inst,$s,$preds)
	}
	set instantiate_res [explorer_instantiate $s $preds]
	debugMsg "instantiate_res($s )($preds)=$instantiate_res"
	if {[lindex $instantiate_res 0]} {
		set new_ev [lindex $instantiate_res 1]
		if {[string compare $si ""] == 0} {
			cache_incr_dump $faut $s [list $new_ev ]
		} else {
			cache_incr_dump $faut $si [list $new_ev ]
		}
		set cache(inst,$s,$preds) $instantiate_res
		incr cache(ntrans) 1
		incr cache(entries)
		incr cache(instmiss)
		debugMsg "cache ($s $preds) instmiss $cache(insthit) $cache(instmiss) $cache(hit) $cache(miss) $cache(entries)"
	}
	return $instantiate_res
}


proc cache_incr_dump {faut s ev} {
	global cache
	
	if {$faut == {}} {
		return
	}
#	set nstates [llength $cache(states)]
#	puts $faut "des($cache(start), $nstates, $cache(ntrans))"
	foreach e $ev {
		set e_nr [lindex $e 0]
		set visible [lindex $e 1]
		set preds_solved  [lindex $e 2]
		set repr [lindex $e 3]
		set preds  [lindex $e 4]
		set freevars  [lindex $e 5]
		set identical [lindex $e 6]
		set time [lindex $e 7]
		set ev_printable $repr
		if {[string compare $identical ""] == 0} {
			puts $faut "($s, \"$repr\", $e_nr)"
		} else {
			puts $faut "($s, \"$repr\", $identical)"
		}
	}
	flush $faut
}
proc cache_dump {faut} {
	global cache
	
	set nstates [llength $cache(states)]
	puts $faut "des($cache(start), $nstates, $cache(ntrans))"
	foreach s [lsort -numerical $cache(states)] {
		cache_incr_dump $faut $s $cache(events,$s)
	}
}
