# =========================================================================
# $Id: menu.tcl,v 1.1.2.29 2008/04/11 11:54:05 belinfan Exp $
# =========================================================================

proc obs_menu { S {useid 0} {id {}}} {
	expansion_empty
	if {$useid} {
		expansion_set_identical $id
	} else {
		expansion_mk_identical_mapping $S
	}
	set identical [expansion_get_identical]
	mr_empty
	reset_menu_verdict
	init_empty
	init_add $S
	set tra_ign [rec_obs_menu $S]
	set transitive [lindex $tra_ign 0]
	set ignorable [lindex $tra_ign 1]
	debugMsg "obs_menu $S += $transitive -= $ignorable == $identical"

# next line now happens on the fly during expansion
#	expansion_mk_identical_mapping $transitive
	return [list [mr_events] $transitive $ignorable $identical]
}

# ==========================================

proc rec_obs_menu { S } {
	# debugMsg "rec_obs_menu start: $S"
	foreach s $S {
		set tra_ign [expand_state $s]
		foreach ss [lindex $tra_ign 0] {
			if {![init_isin $ss]} {
				set transitiveMap($ss) 1
			}
		}
		foreach ss [lindex $tra_ign 1] {
			set ignoreMap($ss) 1
		}
	}
	set transitive [array names transitiveMap]
	set ignorable [array names ignoreMap]
	# debugMsg "rec_obs_menu end: $S += $transitive"
	return [list $transitive $ignorable]
}

# ==========================================

proc init_empty {} {
	global initMap
	catch { unset initMap }
}
proc init_add { S } {
	global initMap
	foreach s $S {
		set initMap($s) 1
		set id [expansion_identical $s]
		if {[string compare $id ""] != 0} {
			set initMap($id) 1
		}
	}
}
proc init_isin { s } {
	global initMap
	if {[info exists initMap($s)] || [info exists initMap([expansion_identical $s])]} {
		return 1
	} else {
		return 0
	}
}

# ==========================================

proc expansion_empty {} {
	global expansion
	debugMsg "expansion_empty"
	catch { unset expansion }
	set expansion(keys) {}
}

proc expansion_mk_identical_mapping { S } {
	global expansion
	foreach s $S {
		set i [mr_identical $s]
		if {[string compare $i ""] != 0} {
			debugMsg "expansion_mk_identical_mapping: $s=$i"
			if {![info exists expansion(identical,$s)]} {
				lappend expansion(keys) $s
			}
			set expansion(identical,$s) $i
		} else {
			debugMsg "expansion_mk_identical_mapping: no for $s"
		}
	}
}

proc expansion_get_identical {} {
	global expansion
	set res {}
	foreach s $expansion(keys) {
		lappend res $s $expansion(identical,$s)
	}
	return $res
}

proc expansion_set_identical {l} {
	global expansion
	foreach {n v}  $l {
		set expansion(identical,$n) $v
		lappend expansion(keys) $n
	}
}

# we replace all states by their 'identical' ones
# this gives us a 'canonical' superstate
proc expansion_canonicalize { S } {
	global expansion
	set canonical {}
	foreach s $S {
		if {[info exists expansion(identical,$s)]} {
			lappend canonical $expansion(identical,$s)
		} else {
			lappend canonical $s
		}
	}
	return $canonical
}
proc expansion_add {s} {
	global expansion
	expansion_mk_identical_mapping $s
	if {[info exists expansion(identical,$s)]} {
		set i $expansion(identical,$s)
		debugMsg "expansion_add: $s=$i"
		set expansion(event,$s) "$s=$i"
		set expansion(event,$i) "$s=$i"
		return $i
	} else {
		debugMsg "expansion_add: $s"
		set expansion(event,$s) $s
		return {}
	}
}
proc in_expansion {s {i ""}} {
	global expansion
	if {[info exists expansion(event,$s)]} {
		debugMsg "in_expansion ($s=$i): $expansion(event,$s)"
		return 1
	} elseif {[info exists expansion(event,$i)]} {
		debugMsg "in_expansion ($s=$i): $expansion(event,$i)"
		return 1
	} else {
		debugMsg "in_expansion ($s=$i): 0"
		return 0
	}
}
proc expansion_identical {s} {
	global expansion
	set i ""
	if {[info exists expansion(identical,$s)]} {
		set i $expansion(identical,$s)
	}
	return $i
}

proc expand_state { s } {
	global primer
	debugMsg "expanding $s"
	set transitive {}
	set ignorable {}
	set identical [expansion_identical $s]
	if {[in_expansion $s $identical]} {
		debugMsg "expanded $s: in_expansion $s identical $identical"
		return [list $transitive $ignorable]
	}
	set s_identical [expansion_add $s]
	set events [cache_explorer_unfold $s $identical $s_identical]
	if {[llength $events] <= 0} {
		warningMsg "empty menu"
		# disable error return for now:
		# return -code error -errorinfo "empty menu"
	}
	debugMsg "expanded $s: $events"
	foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
		set add_suspension($c) 1
		set suspension_predicates($c) {}
	}
	set expandable {}
	set all_internal 1
	set nevents 0
	foreach e $events {
		incr nevents
		set e_nr [lindex $e 0]
		set visible [lindex $e 1]
		set preds_solved  [lindex $e 2]
		set repr [lindex $e 3]
		set attr {}
		if {[regexp {^(.*) \[(.*)\]$} $repr all repr attrs]} {
			set attrlist [fields_split $attrs ,]
			debugMsg "$repr $attrlist"
			foreach f $attrlist {
				set nm_val [field_first $f]
				set nm [lindex $nm_val 0]
				set val [lindex $nm_val 1]
				if {[string compare $val ""] != 0} {
					lappend attr [list $nm $val]
				}
			}
			set e [lreplace $e 3 3 $repr]
		}
		set preds  [lindex $e 4]
		set freevars  [lindex $e 5]
		set identical [lindex $e 6]
		set time [lindex $e 7]
		set ev_printable $repr
		debugMsg "expanded-handling: $e"
		if {[catch {event_kind $ev_printable $visible} kind_chan ]} {
			puts "A_WARNING UnknownIOKind: event=$ev_printable"
			continue
		}
		set kind [lindex $kind_chan 0]
		set chan [lindex $kind_chan 1]
		debugMsg "visibility of $e_nr: $kind $chan"
		switch -- $kind {
			  input -
			  output {
				set all_internal 0
			  	if {$preds_solved} {
					set add_suspension($chan) 0
				} else {
			  		warningMsg "Predicates Unsolved for $kind $e_nr: $preds"
					lappend suspension_predicates($chan) $preds
				}
				mr_insert $kind $chan $e $repr $attr
			} inputsuspension -
			  outputsuspension {
				set all_internal 0
				if {[config_suspension_detection $chan]} {
			  		warningMsg "menu contains explicit suspension event for channel: \"$chan\""
				} else {
			  		if {$preds_solved} {
						set add_suspension($chan) 0
					} else {
			 	 		warningMsg "Predicates Unsolved for $kind $e_nr: $preds"
						lappend suspension_predicates($chan) $preds
					}
					mr_insert $kind $chan $e $repr
				}
			} internal {
				if {! $primer(ioct)} {
				  	if {$preds_solved} {
				  		foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
							set add_suspension($c) 0
						}
					} else {
				  		warningMsg "Predicates Unsolved for $kind $e_nr: $preds"
				  		foreach c [concat [get_channels inputsuspension] [get_channels outputsuspension]] {
							lappend suspension_predicates($chan) $preds
						}
					}
				}
				if {[mr_insert $kind $chan $e $repr]} {
					if {![in_expansion $e_nr $identical]} {
						lappend expandable $e_nr $identical
					} else {
						debugMsg "not re-expanding $e_nr=$identical"
					}
				} else {
						debugMsg "not expanding (because not inserted in menu) $e_nr=$identical"
				}
			} verdict {
				set_menu_verdict $chan
			}
		}
	}
	if {$all_internal && $nevents > 0} {
		lappend ignorable $s
	}
	foreach {e_nr identical} $expandable {
		# debugMsg "transitively expanding $e_nr=$identical"
		set tra_ign [rec_obs_menu [list $e_nr]]
		set transitive [concat $transitive $e_nr [lindex $tra_ign 0]]
		set ignorable [concat $ignorable [lindex $tra_ign 1]]
	}
	foreach k [list inputsuspension outputsuspension] {
		foreach c [get_channels $k] {
			if {$add_suspension($c)} {
				if {[config_suspension_detection $c]} {
					if {[catch {list [suspension_representation $c] } repr]} {
						set repr [list "" ]
					}
					# mr_insert $k $c $s $repr $suspension_predicates
					# NOTE: Should we also copy the predicates and freevars
					# of the transition that reached the current state???
					set susp_s [strip_edge $s]
					set susp_s_identical [strip_edge $s_identical]
					mr_insert $k $c [list $susp_s 1 1 $repr {} {} $susp_s_identical] $repr
				} else {
			  		warningMsg "empty menu (without suspension event) for channel: \"$c\""
				}
			}
		}
	}
	# debugMsg "expand_state $s += $transitive -= $ignorable"
	return [list $transitive $ignorable]
}

proc event_kind {repr {visible 1}} {
	global gates
	global config
	
	debugMsg "event_kind: searching ($repr)"
	if {! $visible} {
		return [list internal ]
	} else {
		set ok_kind_chan [em_find_string $config(emset) $repr]
		if {[lindex $ok_kind_chan 0]} {
			return [lindex $ok_kind_chan 1]
		} else {
			return -code error -errorinfo "unknown event kind"
		}
	}
}

proc event_kind_init {file} {
	global gates
	global delta_event_repr

	config_init
	config_from_file $file
	config_finish
}

proc suspension_representation {chan} {
	return [config_get_suspension_representation $chan]
}

proc get_channels { kind } {
	return [config_get_channels $kind]
}

proc get_kind { channel } {
	return [config_get_iokind $channel]
}

proc is_channel { chan } {
	return [config_is_channel $chan]
}

proc is_iokind { kind } {
	return [config_is_iokind $kind]
}

proc strip_edge { ses } {
	if {[regexp {^(.*[0-9]+)_([^.]*)(\..*)$} $ses all loc edge state]} {
		return "$loc$state"
	} else {
		return $ses
	}
}

proc strip_all_except_state { ses } {
	if {[regexp {^.*\.([^.]*)$} $ses all state]} {
		return "$state"
	} else {
		return $ses
	}
}
