#

proc config_init {li lu delta susp} {
	global config
	debugMsg "config_init ($li)($lu)($delta)($susp)"
	em_init
	catch {unset config}
	set config(channels) {}
	set config(detect,inputsuspension) 0
	set config(detect,outputsuspension) 0
	foreach kind [split $susp ,] {
		switch -- $kind {
			  input {
				set config(detect,inputsuspension) 1
			} output {
				set config(detect,outputsuspension) 1
			} default {
				warningMsg "Unknown iokind for suspension detection: $kind"
			}
		}
	}
}
proc config_finish {} {
	em_compile
}
proc config_cleanup {} {
	catch {unset config}
}

proc config_from_file {filename} {
	if {[catch {open $filename r} fileId]} {
		warningMsg "Cannot open $filename: $fileId"
	} else {
		config_process_file $fileId
		close $fileId
	}
}

proc config_process_file {channel} {
	foreach line [split [read $channel] \n] {
		config_process_line $line
	}
	config_handle_block
}

proc config_process_line {line} {
	global configtmp
	if {[regexp {^\s*$} $line]} {         # empty line
		return
	}
	if {[regexp {^\s*#} $line]} {         # comment line
		return
	}
	if {[regexp {^[^ \t]} $line]} {       # start of new, end of prev block
		config_handle_block
	}
	if {![info exists configtmp]} {
		set configtmp {}
	}
	set configtmp [concat $configtmp [fields_split $line " \t"]]
}

proc config_handle_block {} {
	global configtmp
	global config
	if {![info exists configtmp]} {
		return
	}
	debugMsg "Fields: $configtmp"
	foreach name {channel verdict iokind suspension} {
		set $name [field_value $configtmp $name]
	}
	foreach name {event} {
		set $name [field_value_array $configtmp $name]
	}
	catch {unset configtmp}
	if {([string compare $channel ""] != 0) &&
	    ([string compare $verdict ""] != 0)} {
		warningMsg "ConfigInconsistency: only one allowed of channel, verdict"
	}
	if {[llength $event] <= 0} {
		warningMsg "ConfigError: field missing: event"
	}
	if {[string compare $channel ""] != 0} {
		if {[lsearch $config(channels) $channel] >= 0} {
			warningMsg "ConfigError: duplicate configuration for channel: $channel"
		} else {
			lappend config(channels) $channel
		}
		if {[string compare $iokind ""] == 0} {
			warningMsg "ConfigError: field missing: iokind"
		}
		if {([string compare $iokind "input"] != 0) &&
		    ([string compare $iokind "output"] != 0)} {
			warningMsg "ConfigError: illegal value: iokind should be in (input,output)"
		}
		lappend config(channels,$iokind) $channel
		set config(iokind,$channel) $iokind
		foreach p $event {
			em_add_pattern $p [list $iokind $channel]
		}
		if {[string compare $suspension ""] != 0} {
			if {[regexp {\*} $suspension]} {
				warningMsg "ConfigError: suspension not allowed to contain wildcards: $suspension"
			} else {
				lappend config(channels,[set iokind]suspension) $channel
				em_add_pattern $suspension [list [set iokind]suspension $channel]
				set config(suspension,$iokind,$channel) $suspension
			}
		}
	} elseif {[string compare $verdict ""] != 0} {
		foreach p $event {
			em_add_pattern $p [list verdict $verdict]
		}
		
	} else {
		warningMsg "Config: missing: config block should contain channel or verdict"
	}
}

proc config_get_channels {iokind} {
	global config
	if {[info exists config(channels,$iokind)]} {
		return $config(channels,$iokind)
	} else {
		return {}
	}
}

proc config_get_iokind {chan} {
	global config
	if {[info exists config(iokind,$chan)]} {
		return $config(iokind,$chan)
	} else {
		return ""
	}
}

proc config_get_suspension_representation {channel} {
	global config
	set res ""
	if {[info exists config(iokind,$channel)]} {
		set iokind $config(iokind,$channel)
		if {[info exists config(suspension,$iokind,$channel)]} {
			set res $config(suspension,$iokind,$channel)
		}
	}
	return $res
}

proc config_suspension_detection {iokind} {
	global config
	if {[info exists config(detect,$iokind)]} {
		return $config(detect,$iokind)
	} else {
		return 0
	}
}

proc config_is_channel { chan } {
	global config
	if {[info exists config(iokind,$chan)]} {
		return 1
	} else {
		return 0
	}
}

proc config_is_iokind { kind } {
	global config
	if {[info exists config(channels,$kind)]} {
		return 1
	} else {
		return 0
	}
}
