/***** spin: main.c *****/

/* Copyright (c) 1991,1995 by AT&T Corporation.  All Rights Reserved.     */
/* This software is for educational purposes only.                        */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.att.com          */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "spin.h"
#include "version.h"
#include <signal.h>
#include <stdlib.h>
#include "spiny.h"

extern int	DstepStart, lineno;
extern FILE	*yyin, *tl_out;
extern FILE	*yyout; /* JF: Declaration yyout used in spinlex.c */
extern Symbol	*context;
extern char	*claimproc;

Symbol	*Fname, *oFname;

int	Etimeouts=0;	/* nr timeouts in program */
int	Ntimeouts=0;	/* nr timeouts in never claim */
int	analyze=0, dataflow=0, dumptab=0, has_remote=0;
int	interactive=0, jumpsteps=0, m_loss=0, nr_errs=0;
int	s_trail=0, verbose=0, xspin=0;

static Element	*Same;
static int	IsAsgn = 0, OrIsAsgn = 0;
static char	Operator[] = "operator: ";
static char	Keyword[]  = "keyword: ";
static char	Function[] = "function-name: ";
static char	**add_ltl = (char **)0;

static void	explain(int);

void
alldone(int estatus)
{
	if (xspin && (analyze || s_trail))
	{	if (estatus)
			printf("spin: %d error(s) - aborting\n",
			estatus);
		else
			printf("Exit-Status 0\n");
	}
	exit(estatus);
}

/* AFEB changed return type of main from void to int
 * to get rid of gcc warning
 */
int
main(int argc, char *argv[])
{	Symbol *s;
	int T = (int) time((long *)0);
#ifdef PC
	char out1[] = "_tmp1_";
#else
	char out1[32];
	extern char *tmpnam(char *);
	(void) tmpnam(out1);
#endif
		
	tl_out = stdout; /* JF: Initialization tl_out moved from (tl_buchi.c) */
	yyin   = stdin; /* JF: Initialization yyin moved from (spinlex.c) */
	yyout  = stdout; /* JF: Initialization yyout moved from (spinlex.c) */
	analyze = 1; /* RGV always generate pan files */
	
	while (argc > 1 && argv[1][0] == '-')
	{	switch (argv[1][1]) {
		case 'a': analyze  =  1; break;
/* RGV */
#if 0
		case 'd': dumptab  =  1; break;
		case 'D': dataflow++; break;
		case 'f': add_ltl = (char **) argv;
			  argc--; argv++; break;
		case 'g': verbose +=  1; break;
		case 'i': interactive = 1; break;
		case 'j': jumpsteps = atoi(&argv[1][2]); break;
		case 'l': verbose +=  2; break;
		case 'm': m_loss   =  1; break;
		case 'n': T = atoi(&argv[1][2]); break;
		case 'p': verbose +=  4; break;
		case 'r': verbose +=  8; break;
		case 's': verbose += 16; break;
		case 't': s_trail  =  1; break;
		case 'v': verbose += 32; break;
#endif
		case 'V': printf("%s - version %s (B%s) -\n"
			  "Based on SPIN version %s\n",
			  TanTitle, trojka_version,trojka_build,
			  spin_base_version);
			  alldone(0);
			  break;
#if 0
		case 'X': xspin    =  1;
#endif
#ifndef PC
# ifdef SIGPIPE  /* AFEB added this ifdef */
			  signal(SIGPIPE, alldone); /* not posix... */
# endif
#endif
			  break;
/* RGV */	default : printf("use: spin [-option] ... [-option] file\n");
			  printf("\tnote: file must be the last argument\n");
			  printf("\t-a generate a trojka analyser in tan.c\n");
#if 0
			  printf("\t-d produce symbol-table information\n");
			  printf("\t-D write/write dataflow ");
				printf("(local scalars only)\n");
			  printf("\t-D -D read/write dataflow ");
				printf("(local scalars only)\n");
			  printf("\t-f \"..formula..\"  translate LTL ");
				printf("into never claim\n");
			  printf("\t-g print all global variables\n");
			  printf("\t-i interactive (random simulation)\n");
			  printf("\t-jN skip the first N steps ");
				printf("in simulation trail\n");
			  printf("\t-l print all local variables\n");
			  printf("\t-m lose msgs sent to full queues\n");
			  printf("\t-nN seed for random nr generator\n");
			  printf("\t-p print all statements\n");
			  printf("\t-r print receive events\n");
			  printf("\t-s print send events\n");
			  printf("\t-v verbose, more warnings\n");
			  printf("\t-t follow a simulation trail\n");
#endif
			  printf("\t-V print version number and exit\n");
			  alldone(1);
		}
		argc--, argv++;
	}
	if (argc > 1)
	{	char cmd[128];

		if (add_ltl)
		{
#ifdef PC
#define CPP	"cpp"		/* or: "/gcc/bin/cpp"     */
#define CAT	"type"
			char out2 = "_tmp2_";
#else
#ifdef TROJKACPPSTR
#define CPP TROJKACPPSTR
#else
#define CPP	"/lib/cpp"	/* or: "/usr/ccs/lib/cpp" */
#endif
#define CAT	"cat"
			char out2[32]; (void) tmpnam(out2);
#endif
			sprintf(cmd, "%s %s > %s", CAT, argv[1], out2);
			if (system((const char *)cmd))
			{	(void) unlink((const char *)out2);
				alldone(1);	/* no return */
			}
			if (!(tl_out = fopen(out2, "a")))
			{	printf("spin: cannot append to %s\n", out2);
				alldone(1);
			}
			nr_errs = tl_main(2, add_ltl);	/* in tl_main.c */
			fclose(tl_out);
			sprintf(cmd, "%s -DTROJKA %s > %s", CPP, out2, out1);
			if (system((const char *)cmd))
			{	(void) unlink((const char *)out1);
				(void) unlink((const char *)out2);
				alldone(1);	/* no return */
			}
			(void) unlink((const char *)out2);
		} else
		{	sprintf(cmd, "%s -DTROJKA %s > %s", CPP, argv[1], out1);
			if (system((const char *)cmd))
			{	(void) unlink((const char *)out1);
				alldone(1);	/* no return */
		}	}

		if (!(yyin = fopen(out1, "r")))
		{	printf("cannot open %s\n", out1);
			alldone(1);
		}

		oFname = Fname = lookup(argv[1]);
		if (oFname->name[0] == '\"')
		{	int i = strlen(oFname->name);
			oFname->name[i-1] = '\0';
			oFname = lookup(&oFname->name[1]);
		}
	} else
	{	oFname = Fname = lookup("<stdin>");
		if (add_ltl)
		{	tl_out = stdout;
			exit(tl_main(2, add_ltl));
		}
		printf("%s - version %s (B%S) -\n",
		  TanTitle, trojka_version, trojka_build);
		printf("reading input from stdin:\n");
		fflush(stdout);
	}

	Srand(T);	/* defined in run.c */
	s = lookup("_");	s->type = PREDEF; /* a write only global var */
	s = lookup("_p");	s->type = PREDEF;
	s = lookup("_pid");	s->type = PREDEF;
	s = lookup("_last");	s->type = PREDEF;
	yyparse();
	fclose(yyin);
	if (argc > 1) (void) unlink((const char *)out1);
	alldone(nr_errs);	/* no return */
	return(nr_errs);	/* added by AFEB */
}

int
yywrap(void)	/* dummy routine */
{
	return 1;
}

void
non_fatal(char *s1, char *s2)
{	extern int yychar; extern char yytext[];

	printf("spin: line %3d %s, Error: ", lineno, Fname->name);
	if (s2)
		printf(s1, s2);
	else
		printf(s1);
	if (yychar != -1 && yychar != 0)
	{	printf("	saw '");
		explain(yychar);
		printf("'");
	}
	if (yytext && strlen(yytext)>1)
		printf(" near '%s'", yytext);
	printf("\n"); fflush(stdout);
	nr_errs++;
}

void
fatal(char *s1, char *s2)
{
	non_fatal(s1, s2);
	alldone(1);
}

char *
emalloc(int n)
{	char *tmp;

	if (!(tmp = (char *) malloc(n)))
		fatal("not enough memory", (char *)0);
	memset(tmp, 0, n);
	return tmp;
}

Lextok *
nn(Lextok *s, int t, Lextok *ll, Lextok *rl)
{	Lextok *n = (Lextok *) emalloc(sizeof(Lextok));

	n->ntyp = (short) t;
	if (s && s->fn)
	{	n->ln = s->ln;
		n->fn = s->fn;
	} else if (rl && rl->fn)
	{	n->ln = rl->ln;
		n->fn = rl->fn;
	} else if (ll && ll->fn)
	{	n->ln = ll->ln;
		n->fn = ll->fn;
	} else
	{	n->ln = lineno;
		n->fn = Fname;
	}
	if (s) n->sym  = s->sym;
	n->lft  = ll;
	n->rgt  = rl;
	n->indstep = DstepStart;

	if (t == TIMEOUT) Etimeouts++;

	if (!context)
		return n;
	if (context->name == claimproc)
	{	switch (t) {
		case ASGN:
			printf("spin: Warning, never claim has side-effect\n");
			break;
		case 'r': case 's':
			non_fatal("never claim contains i/o stmnts",(char *)0);
			break;
		case TIMEOUT:
			/* never claim polls timeout */
			Ntimeouts++; Etimeouts--;
			break;
		case LEN: case EMPTY: case FULL:
		case 'R': case NFULL: case NEMPTY:
			/* status bumped to non-exclusive */
			if (n->sym) n->sym->xu |= XX;
			break;
		default:
			break;
		}
	} else if (t == ENABLED || t == PC_VAL)
	{	printf("spin: Warning, using %s outside never-claim\n",
			(t == ENABLED)?"enabled()":"pc_value()");
	}
	return n;
}

Lextok *
rem_lab(Symbol *a, Lextok *b, Symbol *c)
{	Lextok *tmp1, *tmp2, *tmp3;

	has_remote++;
	fix_dest(c, a);	/* in case target is jump */
	tmp1 = nn(ZN, '?',   b, ZN); tmp1->sym = a;
	tmp1 = nn(ZN, 'p', tmp1, ZN);
	tmp1->sym = lookup("_p");
	tmp2 = nn(ZN, NAME,  ZN, ZN); tmp2->sym = a;
	tmp3 = nn(ZN, 'q', tmp2, ZN); tmp3->sym = c;
	return nn(ZN, EQ, tmp1, tmp3);
}

static void
explain(int n)
{
	switch (n) {
	default:	if (n > 0 && n < 256)
				printf("%c' = '", n);
			printf("%d", n);
			break;
	case '\b':	printf("\\b"); break;
	case '\t':	printf("\\t"); break;
	case '\f':	printf("\\f"); break;
	case '\n':	printf("\\n"); break;
	case '\r':	printf("\\r"); break;
	case 'c':	printf("condition"); break;
	case 's':	printf("send"); break;
	case 'r':	printf("recv"); break;
	case '@':	printf("delproc"); break;
	case '?':	printf("(x->y:z)"); break;
	case ACTIVE:	printf("%sactive",	Keyword); break;
	case AND:	printf("%s&&",		Operator); break;
	case ASGN:	printf("%s=",		Operator); break;
	case ASSERT:	printf("%sassert",	Function); break;
	case ATOMIC:	printf("%satomic",	Keyword); break;
	case BREAK:	printf("%sbreak",	Keyword); break;
	case CLAIM:	printf("%snever",	Keyword); break;
	case CONST:	printf("a constant"); break;
	case DECR:	printf("%s--",		Operator); break;
	case D_STEP:	printf("%sd_step",	Keyword); break;
	case DO:	printf("%sdo",		Keyword); break;
	case DOT:	printf("."); break;
	case ELSE:	printf("%selse",	Keyword); break;
	case EMPTY:	printf("%sempty",	Function); break;
	case ENABLED:	printf("%senabled",	Function); break;
	case EQ:	printf("%s==",		Operator); break;
	case FI:	printf("%sfi",		Keyword); break;
	case FULL:	printf("%sfull",	Function); break;
	case GE:	printf("%s>=",		Operator); break;
	case GOTO:	printf("%sgoto",	Keyword); break;
	case GT:	printf("%s>",		Operator); break;
	case IF:	printf("%sif",		Keyword); break;
	case INCR:	printf("%s++",		Operator); break;
	case INIT:	printf("%sinit",	Keyword); break;
	case LABEL:	printf("a label-name"); break;
	case LE:	printf("%s<=",		Operator); break;
	case LEN:	printf("%slen",		Function); break;
	case LSHIFT:	printf("%s<<",		Operator); break;
	case LT:	printf("%s<",		Operator); break;
	case MTYPE:	printf("%smtype",	Keyword); break;
	case NAME:	printf("an identifier"); break;
	case NE:	printf("%s!=",		Operator); break;
	case NEG:	printf("%s! (not)",	Operator); break;
	case NEMPTY:	printf("%snempty",	Function); break;
	case NFULL:	printf("%snfull",	Function); break;
	case NON_ATOMIC: printf("sub-sequence"); break;
	case OD:	printf("%sod",		Keyword); break;
	case OF:	printf("%sof",		Keyword); break;
	case OR:	printf("%s||",		Operator); break;
	case O_SND:	printf("%s!!",		Operator); break;
	case PC_VAL:	printf("%spc_value",	Function); break;
	case PNAME:	printf("process name"); break;
	case PRINT:	printf("%sprintf",	Function); break;
	case PROCTYPE:	printf("%sproctype",	Keyword); break;
	case RCV:	printf("%s?",		Operator); break;
	case R_RCV:	printf("%s??",		Operator); break;
	case RSHIFT:	printf("%s>>",		Operator); break;
	case RUN:	printf("%srun",		Operator); break;
	case SEP:	printf("token: ::"); break;
	case SEMI:	printf(";"); break;
	case SND:	printf("%s!",		Operator); break;
	case STRING:	printf("a string"); break;
	case TIMEOUT:	printf("%stimeout",	Keyword); break;
	case TYPE:	printf("data typename"); break;
	case TYPEDEF:	printf("%stypedef",	Keyword); break;
	case XU:	printf("%sx[rs]",	Keyword); break;
	case UMIN:	printf("%s- (unary minus)", Operator); break;
	case UNAME:	printf("a typename"); break;
	case UNLESS:	printf("%sunless",	Keyword); break;
	}
}

static int
used_here(Symbol *s, Lextok *n)
{	int res = 0;

	if (!n) return 0;
#ifdef DEBUG
	{	int oln; Symbol *ofn;
		printf("	used_here %s -- ", context->name);
		oln = lineno; ofn = Fname;
		comment(stdout, n, 0);
		lineno = oln; Fname = ofn;
		printf(" -- %d:%s\n", n->ln,n->fn->name);
	}
#endif
	if (n->sym == s) res = (IsAsgn || n->ntyp == ASGN)?2:1;
	if (n->ntyp == ASGN)
		res |= used_here(s, n->lft->lft);
	else
		res |= used_here(s, n->lft);
	if (n->ntyp == 'r')
		IsAsgn = 1;
	res |= used_here(s, n->rgt);
	if (n->ntyp == 'r')
		IsAsgn = 0;
	return res;
}

static int
used_later(Symbol *s, Element *t)
{	int res = 0;

	if (!t || !s)
		return 0;
	if (t->status&CHECK2)
	{
#ifdef DEBUG
		printf("\t%d used_later: done before\n", t->Seqno);
#endif
		return (t->Seqno == Same->Seqno) ? 4 : 0;
	}

	t->status |= CHECK2;

#ifdef DEBUG
	{	int oln; Symbol *ofn;
		printf("\t%d ->%d used_later %s -- ",
			t->seqno, (t->nxt)?t->nxt->seqno:-1,
			context->name);
		oln = lineno; ofn = Fname;
		comment(stdout, t->n, 0);
		lineno = oln; Fname = ofn;
		printf(" -- %d:%s\n", t->n->ln, t->n->fn->name);
	}
#endif
	if (t->n->ntyp == GOTO)
	{	Element *j = target(t);
#ifdef DEBUG
		printf("\t\tjump to %d\n", j?j->Seqno:-1);
#endif
		res |= used_later(s, j);
		goto done;
	}

	if (t->n->sl && ! t->sub)	/* d_step or (non-) atomic */
	{	SeqList *f;
		for (f = t->n->sl; f; f = f->nxt)
		{	f->this->last->nxt = t->nxt;
			res |= used_later(s, f->this->frst);
		}
	} else if (t->sub)	/* IF or DO */
	{	SeqList *f;
		for (f = t->sub; f; f = f->nxt)
			res |= used_later(s, f->this->frst);
	} else
	{	res |= used_here(s, t->n);
	}
	if (!(res&3)) res |= used_later(s, t->nxt);
done:
	t->status &= ~CHECK2;
	return res;
}

static void
varused(Lextok *t, Element *u, int isread)
{	int res = 0;

	if (!t || !t->sym)		return;
	if (dataflow == 1 && isread)	return;

	res = used_later(t->sym, u);

	if ((res&1)
	||  (isread && res&4))
		 return;	/* followed by at least one read */

	printf("%s: %3d: ", 
		Same->n->fn->name,
		Same->n->ln);

	if (t->sym->type == STRUCT)
		full_name(stdout, t, t->sym, 1);
	else
		printf("%s", t->sym->name);

	printf(" -- (%s)", (isread)?"read":"write");

	if (!res)	printf(" none");
	if (res&2)	printf(" write");
	if (res&4)	printf(" same");
	printf("\n");

}

static void
varprobe(Element *parent, Lextok *n, Element *q)
{
	if (!n) return;

	Same = parent;

	/* can't deal with globals, structs, or arrays */
	if (n->sym
	&&  n->sym->context
	&&  n->sym->nel == 1
	&&  n->sym->type != STRUCT
	&&  n->ntyp != 'p'
	&&  n->ntyp != 'q'
	&&  n->ntyp != PRINT)
		varused(n, q, (!OrIsAsgn && n->ntyp != ASGN));

	if (n->ntyp == ASGN)
		varprobe(parent, n->lft->lft, q);
	else
		varprobe(parent, n->lft, q);

	if (n->ntyp == 'r')
		OrIsAsgn = 1;

	varprobe(parent, n->rgt, q);

	if (n->ntyp == 'r')
		OrIsAsgn = 0;
}

#if 0
#define walkprog	varcheck
#else
#define Varcheck	varcheck
#endif

void
Varcheck(Element *e, Element *nx)
{	SeqList *f;

	if (!dataflow || !e || e->status&CHECK1)
		return;
#ifdef DEBUG
	{	int oln; Symbol *ofn;
		printf("%s:%d -- ", context->name, e->Seqno);
		oln = lineno; ofn = Fname;
		comment(stdout, e->n, 0);
		lineno = oln; Fname = ofn;
		printf(" -- %d:%s\n", e->n->ln, e->n->fn->name);
	}
#endif
	e->status |= CHECK1;

	if (e->n->ntyp == GOTO)
	{	Element *ef = target(e);
		if (ef) varcheck(ef, ef->nxt);
		goto done;
	} else if (e->n->sl && !e->sub)	/* d_step or (non)-atomic */
	{	for (f = e->n->sl; f; f = f->nxt)
		{	f->this->last->nxt = nx;
#ifdef DEBUG
			printf("\tseq %d->%d\n",
			f->this->last->seqno, nx?nx->seqno:-1);
#endif
			varcheck(f->this->frst, nx);
			f->this->last->nxt = 0;
		}
	} else if (e->sub && e->n->ntyp == IF)	/* if */
	{	for (f = e->sub; f; f = f->nxt)
			varcheck(f->this->frst, nx);
	} else if (e->sub && e->n->ntyp == DO)	/* do */
	{	for (f = e->sub; f; f = f->nxt)
			varcheck(f->this->frst, e);
	} else
	{	varprobe(e, e->n, e->nxt);
	}
	{	Element *ef = huntele(e->nxt, e->status);
		if (ef) varcheck(ef, ef->nxt);
	}
done:
	/* e->status &= ~CHECK1 */ ;
}

static void
nested(int n)
{	int i;
	for (i = 0; i < n; i++)
		printf("\t");
}

static void
walkprog(Element *e, Element *nx)
{	SeqList *f;
	static int Nest=0; int oln;

	if (!dataflow) return;
	if (!e)
	{	nested(Nest);
		printf("nil\n");
		return;
	}

	nested(Nest);
	printf("%4d,%4d, %s:%d -- ",
		e->status, lineno,
		context->name, e->Seqno);
	oln = lineno;
	comment(stdout, e->n, 0);
	lineno = oln;
	printf(" -- %d:%s\n", e->n->ln, e->n->fn->name);

	if (e->status&CHECK1)
	{	nested(Nest);
		printf("seenbefore\n");
		return;
	}

	e->status |= CHECK1;

	if (e->n->ntyp == GOTO)
	{	Element *ef = target(e);
		if (ef) walkprog(ef, ef->nxt);
	} else if (e->n->sl && !e->sub)	/* ATOMIC, NON_ATOMIC, D_STEP */
	{	int cnt;

		for (f = e->n->sl, cnt=1; f; f = f->nxt, cnt++)
		{	Nest++;
			nested(Nest);
			printf("---a>%d:\n", cnt);
#ifdef DEBUG
			printf("\tseq %d->%d\n",
			f->this->last->seqno, nx?nx->seqno:-1);
			f->this->last->nxt = nx;
#endif
			walkprog(f->this->frst, nx);
		/*	f->this->last->nxt = 0;		*/
			Nest--;
		}
	} else if (e->sub && e->n->ntyp == IF)
	{	int cnt;
		for (f = e->sub, cnt=1; f; f = f->nxt, cnt++)
		{	Nest++;
			nested(Nest);
			printf("---s>%d:\n", cnt);
			walkprog(f->this->frst, nx);
			Nest--;
		}
	} else if (e->sub && e->n->ntyp == DO)
	{	int cnt;
		for (f = e->sub, cnt=1; f; f = f->nxt, cnt++)
		{	Nest++;
			nested(Nest);
			printf("---s>%d:\n", cnt);
			walkprog(f->this->frst, e);
			Nest--;
		}
	}
	{	Element *ef = huntele(e->nxt, e->status);
		if (ef) walkprog(ef, ef->nxt);
	}
	e->status &= ~CHECK1;
}
