#! @TCLSH@

# seems this works better with tclsh than with expect on windows
#
# $Id: anifsm.sh,v 1.1.2.62 2008/07/10 23:44:53 belinfan Exp $
#
# (c) 2004-2005 Axel.Belinfante@cs.utwente.nl
# Feel free to use, update or redistribute as long as no monies are exchanged.
# Please report fixes/improvements back to me.
#
# To use, make one change above:
#   - replace /usr/bin/expect by the path to your tclsh or expect interpreter
#   - make sure that the companion script anifsmsrv can be found
#      (is in your PATH)
#   - make sure that the utility script expnohup can be found
#      (is in your PATH)
#
# NOTE: this script only needs tcl(sh); it does not use expect
# features. The only reason that we use the expect interpreter
# here is that the configure script of TorX finds the location of
# expect already for us, and does not look for tclsh.

# === start clientsrv.tcl ===== {
proc find_user {} {
	global env
	global tcl_platform

	if {[info exists tcl_platform(user)]} {
		set user $tcl_platform(user)
	} elseif {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} elseif {[info exists env(USERNAME)]} {
		set user $env(USERNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_tmp {} {
	global env
	return /tmp
	# Stupid temp directory
	set S(tmp) [pwd]
	if {[file isdirectory "c:/temp"]} { set S(tmp) "c:/temp"}
	if {[file isdirectory "/tmp"]} { set S(tmp) "/tmp"}
	catch {set S(tmp) $env(TRASH_FOLDER)}           ;# Macintosh(?)
	catch {set S(tmp) $env(TMPDIR)}
	catch {set S(tmp) $env(TMP)}
	catch {set S(tmp) $env(TEMP)}
	return $S(tmp)
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set dname "$pfx-$user-$display"
	regsub -all : $dname _ dname
	set fname [file join $dir $dname]
	return $fname
}

proc get_address { fname do_warn} {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			set words [split $line]
			set ipNr [lindex $words 0]
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			return "tcp!$ipNr!$portNr"
		}
		close $fid
	} elseif {$do_warn} {
		warn stderr "$fid"
		return ""
	}
}

proc try_connect { address do_warn} {
	set words [split $address !]
	if {[llength $words] == 3 && [string compare [lindex $words 0] tcp] == 0} {
		set host [lindex $words 1]
		set port [lindex $words 2]
		if {[string compare $host 0.0.0.0] == 0} {
			set host 127.0.0.1
		}
		if {![catch {socket $host $port} sock]} {
			return $sock
		} else {
			warn stderr "could not connect tcp!$host!$port: $sock"
		}
	} elseif {$do_warn} {
		warn stderr "could not parse address $address"
	}
	return 0
}
# === end clientsrv.tcl ===== }
# === start process.tcl ===== }
# initially we tried to run 'exec -- tclkit bla -s &' but somehow
# on windows # the '-s' did not arrive with the program that
# we exec-ed
# That's why we tried using '-server' intead of '-s', which does work.


proc kill {p} {
	global tcl_platform

	msg stderr "cleaning up: kill $p"
	set tp $tcl_platform(platform)
	switch -- $tp {
	    unix {
		catch {exec kill $p}
		msg stderr "cleaning up: kill -9 $p"
		catch {exec kill -9 $p}
	  } windows -
	    macintosh {
		msg stderr "don't know how to kill $p (on platform $tp)"
	  } default {
		msg stderr "don't know how to kill $p (unknown platform $tp)"
	  }
	}
}

proc findprocess {pid p} {
	global tcl_platform

	set found 0
	set tp $tcl_platform(platform)
	warn stdout "findprocess $pid $p  $tp"
	switch -- $tp {
	    unix {
		if {[catch {open "|ps -ef"} pidsfd]} {
			warn stdout "could not run: ps -ef"
			exit 1
		} else {
			while {[gets $pidsfd line] >= 0} {
#				warn stdout "findprocess $pid $p  $tp line $line"
#				debug "$pid: looking for $p in: $line"
				if {[lindex [split $line] 1] == $p}  {
#					warn stdout "found running server pid: $p"
					set found 1
					break
				}
			}
			# try to catch: child killed: write on pipe with no readers
			catch {close $pidsfd}
		}
	  } windows {
#		warn stdout "findprocess windows"
		if {![catch {open "|tasklist"} pidsfd]} {
#			warn stdout "findprocess windows in tasklist"
			while {[gets $pidsfd line] >= 0} {
				# debug 1 "$pid: looking for $p in: $line"
				if {[lindex [split $line] 2] == $p}  {
#					warn stdout "tasklist found running server pid: $p"
					set found 1
					break
				}
			}
			close $pidsfd
		} elseif {![catch {open "|tlist"} pidsfd]} {
#			warn stdout "findprocess windows in tlist"
			while {[gets $pidsfd line] >= 0} {
				# debug 1 "$pid: looking for $p in: $line"
				set nr [lindex [split $line] 0]
				set nr2 [expr int($nr - pow(2,32)]
				if {$nr == $p || $nr2 == $p}  {
#					warn stdout "tlist found running server pid: $p"
					set found 1
					break
				}
			}
			close $pidsfd
		} else {
			# should try tlist on other windows?
			warn stdout "could not run: tasklist"
			exit 1
		}
	  } macintosh {
		warn stdout "don't know how to kill $p (on platform $tp)"
	  } default {
		warn stdout "don't know how to kill $p (unknown platform $tp)"
	  }
	}
	warn stdout "findprocess $pid $p: $found"
	return $found
}

# === end process.tcl ===== }


set debug 0
set G(prog) anifsm
set G(srvprog) anifsmsrv
set G(fnamepart)  anifsm

set prog anifsm

# we use output to make sure that everything is flushed!
# we do this to avoid deadlocks due to buffers being full...
proc warn { fid s } {
	global PID
	global G
	catch {
		puts $fid "$G(sprog):$PID warning: $s"
		flush $fid
	}
}
proc output {chan msg} {
	catch { puts $chan [string trim $msg] }
	catch { flush $chan }
}

proc fatal {s {code {0}}} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
	cleanup_exit $code
}
proc message {s} {
	global prog
	set s [string trim $s]
	output stdout "$prog: $s"
}
proc warning {s} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
}

proc debug {s} {
	global prog
	global debug
	if {$debug} {
		set s [string trim $s]
		output stderr "$prog: $s"
	}
}

proc cleanup_exit {{code {0}}} {
	global state
	if {[info exists state(stderr)]} {
		eval fconfigure stderr $state(stderr)
	}
	if {[info exists state(stdout)]} {
		eval fconfigure stdout $state(stdout)
	}
	if {[info exists state(stdin)]} {
		eval fconfigure stdin $state(stdin)
	}
	exit $code
}

proc configure_fids {socket} {
	fconfigure $socket -buffering line -blocking 0
# not for windows: ?
# \
#		                   -translation {binary binary}

}

proc filevent_fids {socket remoteHost remotePort waitRemoteClose} {
	global  forever
	fileevent $socket readable \
	    	[list getRemote $remoteHost $remotePort $socket]
	fileevent stdin readable \
	 	[list sendRemote $remoteHost $remotePort $socket $waitRemoteClose]
	vwait forever
}


proc getRemote {remoteHost remotePort sock} {
	if {[eof $sock]} {
		catch {close $sock }
		catch { close stdout }
 		cleanup_exit
	} else {
		set read [gets $sock line]
		debug "getRemote tcp!$remoteHost!$remotePort read: $read"
		if {$read < 0} {
 		} else {
			debug "getRemote tcp!$remoteHost!$remotePort: $line"
			output stdout $line
		}
	}
}
proc sendRemote {remoteHost remotePort sock waitRemoteClose} {
	if {[eof stdin]} {
		output $sock -eof
		debug "eof stdin"
		if {! $waitRemoteClose} {
			catch {close $sock}
			debug "sendRemote closed tcp!$remoteHost!$remotePort"
			cleanup_exit
		} else {
			return
		}
	} else {
		set read [gets stdin line]
		debug "sendRemote tcp!$remoteHost!$remotePort read: $read"
		if {$read < 0} {
		} else {
			debug "sendRemote tcp!$remoteHost!$remotePort: $line"
			output $sock $line
		}
	}
}

proc main_loop {sock waitRemoteClose argstring} {
	configure_fids $sock
	set remote [fconfigure $sock -peername]
	set remoteHost [lindex $remote 0]
	set remotePort [lindex $remote 2]
	puts $sock $argstring
	filevent_fids $sock $remoteHost $remotePort $waitRemoteClose
}

proc usage {} {
	warning "usage: $G(prog) \[-r\] \[-m mcastid\] \[-t title\] \[-k key\] \[-s srvaddr \] \[file.dot\]"
		exit 1
}

proc main {} {
	global argv
	global env
	global pid
	global PID
	global G

	set waitRemoteClose 0

	set reuse ""
	set mcast ""
	set title ""
	set key ""
	set srvaddr ""
	set files {}
	set exit 0
	set errors 0
	set pid [pid]
	set PID $pid

	set argstring ""
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast "-m $env(TORXMCASTID)"
	}

	set args $argv
	
#	set joinedargs [join $argv ")(" ]
#	warning "command lin args: ($joinedargs)"

	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse "-r"
		  } -m {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
					set errors 1
				} else {
					set mcast "-m [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
					set errors 1
				} else {
					set title "-t {[lindex $args 0]}"
					set args [lrange $args 1 end]
				}
		  } -k {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
					set errors 1
				} else {
					set key "-k [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -s {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
					set errors 1
				} else {
					set srvaddr [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -exit {
		  		set exit 1
		  } -* {
		  		warning "unknown flag: $a"
				set errors 1
		  } default {
		  		lappend files $a
		  }
		}
	}

	if {$errors} {
		usage
	}
	set host [info hostname]
	set display [find_display $host]
	set user [find_user]
	set tmp [find_tmp]
	set fname [mk_filename $tmp $G(fnamepart) $user $display]
	# puts stderr "fname=$fname"
	set B(interface,name) $fname
	set B(pidsfile,name) "$fname.pid"
	set B(statusfile,name) "$fname.stat"

	if {[string compare $srvaddr ""] == 0} {
		set srvaddr [get_address $fname 0]
	}
	set sock [try_connect $srvaddr 0]

	if {$exit} {
		if {[string compare $sock 0] != 0} {
			puts $sock -exit
			exit
		}
	} else {
		set ff {}
		foreach file $files {
			if {[string compare $file "-"] != 0 && [string compare [file pathtype $file] absolute] != 0} {
				set file [file join [pwd] $file]
			}
			lappend ff $file
		}
# 		set argstring "-args $reuse $mcast $title $key [list $file]"
		set argstring "-args $reuse $mcast $title $key $ff"
		# puts stderr "sending $argstring"

		if {[string compare $sock 0] != 0} {
#puts  "bla4 $sock"
			main_loop $sock $waitRemoteClose $argstring
		} else {
#puts  "bla4 $sock"
			if {[catch {open $B(statusfile,name) a} fid]} {
				puts  "cannot open (append) $B(statusfile,name): $fid"
				exit 1
			} else {
				puts $fid "starting $host $pid"
				close $fid
			}
			if {[catch {open $B(statusfile,name) r} fid]} {
				puts  "cannot open (read) $B(statusfile,name)"
				exit 1
			} elseif {[gets $fid line] < 0} {
				puts  "cannot read $B(statusfile,name)"
				exit 1
			} else {
				close $fid
				#puts  "read $B(statusfile,name): $line"
				set words [split $line]
				set stat [lindex $words 0]
				set h [lindex $words 1]
				set p [lindex $words 2]
				if {[string compare $stat starting] == 0 &&
				    [string compare $host $h] == 0 &&
				    ( [string compare $pid $p] == 0 ||
				      ![findprocess $pid $p])} {
					# puts  "starting  $G(srvprog)"
					start  [list $G(srvprog) $pid]
				} else {
					puts  "stat $stat (starting) h $h ($host) p $p ($pid)"
				}
				set i 0
				set maxtries 150
				while {$i < $maxtries} {
					# puts  "trying to connect $i)"
					if {$i == ($maxtries - 1)} {
						set sock [try_connect [get_address $fname 1] 1]
					} else {
						set sock [try_connect [get_address $fname 0] 0]
					}
					if {[string compare $sock 0] != 0} {
						# puts "connected $sock"
						main_loop $sock $waitRemoteClose $argstring
					}
					after 100
					incr i
				}
			}
		}
	}
}

proc start { argv } {
	exec sh -c "expnohup $argv &" &
}


main
exit
