#! /bin/sh
#-*-tcl-*-
# the next line restarts using wish \
exec @WISH@ "$0" -- ${1+"$@"}

# we deal here with multiple progressbar animation runs,
# in multiple windows.
# we use 'nr' to refer to a window
# we use 'id' to refer to an animation run

wm withdraw .

set PID [pid]:

set color(background) white
set quit 0

set debug 0

set id 0
set ncanvas 0
set windows {}
set reuseQ {}

proc ldelete {l e} {
	 set ix [lsearch -exact $l $e]
	 if {$ix >= 0} {
	 	return [lreplace $l $ix $ix]
	 } else {
	 	return $l
	 }
}

proc isinteger { s } {
	# ideal implementation: return [string is integer -strict $s]
	# but that does not work in older wish versions
	if {[regexp "^\[-+\]?\[0-9\]\[0-9\]*$" $s]} {
		return 1
	} else {
		return 0
	}
}
proc isdouble { s } {
	# ideal implementation: return [string is double -strict $s]
	# but that does not work in older wish versions
	if {[regexp "^\[-+\]?\[0-9\]\[0-9\]*(\.\[0-9\]\[0-9\]*)?(\[eE\]\[-+\]?\[0-9\]\[0-9\]*)?$" $s]} {
		return 1
	} else {
		return 0
	}
}

proc find_reusable {id ofid} {
	global reuseQ

	debug 1 "find_reusable $id : ($reuseQ) {"
	set res ""
	foreach w_nr $reuseQ {
		upvar #0 W_[set w_nr](id) w_id
		upvar #0 G_[set w_id](ieof) w_ieof
		if {[string compare $id $w_id] == 0} {
			debug 1 "find_reusable $id: skipping $w_nr: id $id == w_id $w_id"
		} elseif {$w_ieof} {
			debug 1 "find_reusable $id: found $w_nr (containing $w_id))"
			set reuseQ [ldelete $reuseQ $w_nr]
			set res $w_nr
			break
		} else {
			debug 1 "find_reusable $id: skipping $w_nr: no eof seen"
		}
	}
	debug 1 "find_reusable $id: ($reuseQ) } ($res)"
	return $res
}

proc getcanvas {id reuse ofid} {
	debug 1 "getcanvas ($id) ($reuse)($ofid) {"
	set nr [find_reusable $id $ofid]
	if {[string compare $nr ""] != 0} {
		upvar #0 W_[set nr](id) w_id
		upvar #0 S_[set w_id](nr) w_nrs
		set res_nr [reusecanvas $w_nrs $id $reuse]
		# assert($res_nr == $nr)
	} else {
		set res_nr [mkcanvas $id $reuse]
	}
	debug 1 "getcanvas ($id) ($reuse)($ofid)($res_nr) }"
	return $res_nr
}

proc reusecanvas {nrs id reuse} {

	debug 1  "reusecanvas ($nrs) ($id) ($reuse) {"
	foreach nr $nrs {
		upvar #0 W_[set nr](doReuse) doReuse
		upvar #0 W_[set nr](id) w_id
		upvar #0 W_[set nr](c) c
		upvar #0 W_[set nr](t) t
		upvar #0 W_[set nr](step) step
		upvar #0 S_[set id](nr) w_nrs
		upvar #0 G_[set id](title) title
		global zfact

		debug 1  "reusecanvas ($nrs) ($id) ($reuse): $nr"
		lappend w_nrs $nr

		destroy_contents $nr
#		$c delete all
		bind $t <Configure> {}

		set w_id $id
		set doReuse $reuse
		set step ""


		wm title $t $title
		wm deiconify $t
		raise $t
	}
	update_reuse [lindex $nrs 0]
	debug 1  "reusecanvas ($nrs) ($id) ($reuse) }"

	return $nrs
}


proc mkcanvas {id reuse} {
	global color
	global ncanvas
	global windows
	global SCROLLBARWIDTH
	upvar #0 S_[set id](nr) w_nr
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](title) title
	upvar #0 W_[set ncanvas](doReuse) doReuse
	upvar #0 W_[set ncanvas](busy) busy
	upvar #0 W_[set ncanvas](destroying) destroying
	upvar #0 W_[set ncanvas](id) w_id
	upvar #0 W_[set ncanvas](c) c
	upvar #0 W_[set ncanvas](t) t
	upvar #0 W_[set ncanvas](step) step

	# we use busy and destroying to synchronise with the destroy
	# to avoid trying to write to a window that is just destroyed
	set busy 0
	set destroying 0

	set nr $ncanvas
	incr ncanvas

	lappend w_nr $nr
	set w_id $id
	set doReuse $reuse

	set step "0"

	set t [toplevel .f$nr -background $color(background)]
	lappend windows $t

	# make frames first, below everything else
	frame $t.f
	frame $t.fc \
		-background  $color(background) \
		-borderwidth 0 \
		-highlightthickness 0
	frame $t.fq -borderwidth 2 -relief ridge
	frame $t.fn -borderwidth 2 -relief ridge
	frame $t.bs

	set c [progressbar $t.c]

	button $t.bq -text "Quit" \
		-padx 2 \
		-command "quit $nr"

	button $t.bc -text "Close" \
		-padx 2 \
		-command "mydestroy $nr"

	checkbutton $t.br -text "Reuse" \
		 -relief raised \
		 -variable W_[set nr](doReuse) \
		 -command "update_reuse $nr"
	update_reuse $nr

	label $t.l -text "Step:"
	entry $t.t \
		 -width 5 \
		 -textvariable W_[set nr](step)
	bind $t.t <Return> "showq $nr 1 \$W_[set nr](step)"
	button $t.bsp -image up \
		-command "step_inc $nr +1"
	button $t.bsm -image down \
		-command "step_inc $nr -1"

	pack append $t.fq \
		$t.bq {right } \
		$t.bc {right } \
		$t.br {right fillx filly }

#		$t.bl {right fillx filly  } \
#
	pack append $t.bs \
		$t.bsp {top  fillx filly expand } \
		$t.bsm {bottom  fillx filly expand } \

	pack append $t.fn \
		$t.bs {right  fillx filly } \
		$t.t {right fillx filly expand } \
		$t.l {left  fillx filly }

	pack append $t.f \
		$t.fq {right } \
		$t.fn {right  fillx filly expand}

	pack append $t.fc \
		$c {top expand fill}

	pack append $t \
		$t.f {bottom fillx} \
		$t.fc {top expand fill}

	wm protocol $t WM_DELETE_WINDOW "mydestroy $nr"
	wm protocol $t WM_SAVE_YOURSELF "mydestroy $nr"
	wm title $t $title
	bind $t <q>  "quit $nr"
	bind $t <Q>  "quit $nr"
	bind $c <1> "next $nr"
	bind $c <3> "prev $nr"


	tkwait visibility $c
	set geo [wm geometry $t]
	if {[regexp {^([0-9]+)x([0-9]+).*} $geo match w h]} {
		wm minsize $t $w $h
	}
	bind $t <Configure>  "$c redisplay"
	return $nr

}

proc update_reuse {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](doReuse) doReuse
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) w_nrs
	global reuseQ

	debug 1 "update_reuse $nr { doReuse $doReuse ieof $ieof reuseQ $reuseQ"
	foreach w_nr $w_nrs {
		upvar #0 W_[set w_nr](doReuse) w_doReuse
		set w_doReuse $doReuse
		debug 1 "update_reuse $w_nr : doReuse $doReuse ieof $ieof reuseQ $reuseQ"
		 if {$ieof && $doReuse && [lsearch -exact  $reuseQ $w_nr] < 0} {
		 	lappend reuseQ $w_nr
		 } elseif {!$doReuse} {
		 	set reuseQ [ldelete $reuseQ $w_nr]
		 }
		 if {$ieof} {
		 	.f$w_nr.br configure -state normal
		 } else {
		 	.f$w_nr.br configure -state disabled
		 }
	}
	debug 1 "update_reuse $nr } reuseQ $reuseQ"
}

proc step_inc {nr inc} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](c) c
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](sprev) sprev
	upvar #0 G_[set id](Q) Q
	
	if {([string compare $scur ""] != 0) &&
	    ([isinteger $scur]) &&
	    ([expr $scur $inc] >= 0) && ([expr $scur $inc] <= [llength $Q])} {
		set idx $scur
	    	incr idx $inc
	    	showq $nr 1 $idx [expr $idx -1]
	}
}

proc prev {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](scur) scur
	set idx $scur
	showq $nr 1 [expr $idx -1] [expr $idx -2] 
}

proc next {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](scur) scur
	set idx $scur
	showq $nr 1 [expr $idx +1] $idx 
}

proc progressbar {w args} {
	array set opt {
		-width 256 -height 16 -relief sunken -borderwidth 1 -style bar
	}
	array set opt $args
	set style $opt(-style)
	unset      opt(-style)  ;# don't let the canvas get this
	eval canvas $w [array get opt]
	set h [$w cget -height] 
	switch -- $style {
		bar {
			$w create rect 1 1 1 [expr {$h-1}] -fill CornflowerBlue \
				-tags {bar status}
		} 
		circle {
			$w create arc  1 1 [expr {$h-1}] [expr {$h-1}] -start 90 \
				-fill CornflowerBlue -tags {arc status} -extent 0 
		}
	}
	$w create text [expr [$w cget -width] / 2] [expr $h / 2] -tags txt
	rename $w _$w
	proc $w {args} {
		global cur
		set w [lindex [info level 0] 0]
		debug 2 "$w $args"
		if {[lindex $args 0] == "set"} {
			set n [expr int([lindex $args 1])]
			set t [lindex $args 2]
			set d [lindex $args 3]
			set r [lindex $args 4]
			set txt "$n % ($d / $t) remains: $r"
			$w itemconf txt -text $txt
			$w draw $n
			set color [color:rgb $n]
			$w itemconf status -fill $color -outline $color
		} elseif {[lindex $args 0] == "draw"} {
			set cur($w,args) $args
			set n [lindex $args 1]
			set dx [winfo width $w]
			set tbbox [$w bbox txt]
			set tdx [expr 20 + [lindex $tbbox 2] -  [lindex $tbbox 0]]
			debug 2 "tdx=$tdx dx=$dx tbbox=$tbbox"
			if {$tdx > $dx} {
				$w configure -width $tdx
				update idletasks
				set geo [wm geometry [winfo toplevel $w]]
				if {[regexp {^([0-9]+)x([0-9]+).*} $geo match gdx gdy]} {
					debug 2 "dx=$tdx dx=$dx gdx=$gdx gdy=$gdy"
					wm minsize [winfo toplevel $w] $gdx $gdy
				}
			}
			set dx [winfo width $w]  
			set dy [winfo height $w]
			$w coords txt [expr $dx / 2] [expr $dy / 2]
			$w coords bar 1 1 [expr $n * $dx / 100] [expr $dy - 1]
			$w itemconf arc -extent [expr $n*-3.599]
		} elseif {[lindex $args 0] == "redisplay"} {
			if {[info exists cur($w,args)] && [lindex $cur($w,args) 0] == "draw"} {
				eval $w $cur($w,args)
			}
		} else {
			eval _$w $args 
		}
	}
	return $w
}
proc color:rgb {n} {
	# map 0..100 to a red-yellow-green sequence
	set n     [expr {$n < 0? 0: $n > 100? 100: $n}]
	set red   [expr {$n > 75? 60 - ($n * 15 / 25) : 15}]
	set green [expr {$n < 50? $n * 15 / 50 : 15}] 
	format    "#%01x%01x0" $red $green
}   

proc stopcolor {nr} {
	upvar #0 W_[set nr](c) c
	$c itemconf status -fill CornflowerBlue -outline CornflowerBlue
}

proc showcounter {nr cmd n t d r} {
	upvar #0 W_[set nr](c) c
	debug 1 "showcounter $nr $cmd $n $t $d $r"
	$c set $n $t $d $r
	if {[string compare $cmd stop] == 0} {
		stopcolor $nr
	}
}

proc setcounter {nr n t d r} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](Q) Q
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](cur_n) cur_n
	upvar #0 G_[set id](cur_t) cur_t
	upvar #0 G_[set id](cur_d) cur_d
	upvar #0 G_[set id](cur_r) cur_r

	debug 1 "setcounter nr=$nr id=$id scur=$scur len_Q=[llength $Q]"

	set cur_n $n
	set cur_t $t
	set cur_d $d
	set cur_r $r
	if {$scur == [llength $Q]} {
		showcounter $nr cur $n $t $d $r
	} else {
		debug 2 "scur=$scur len=[llength $Q]"
	}
}

proc countdown {id sec} {
	upvar #0 S_[set id](nr) nr
	upvar #0 W_[set nr](c) c
	upvar #0 G_[set id](cur_n) cur_n
	upvar #0 G_[set id](cur_t) cur_t
	upvar #0 G_[set id](cur_d) cur_d
	upvar #0 G_[set id](cur_r) cur_r
	upvar #0 G_[set id](cur_c) cur_c
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](Q) Q
	upvar #0 W_[set nr]_afters afters

	debug 1 "countdown ($id) ($sec)"
	set fact 1000.0
	set count  [expr $sec * $fact]
	cancel $nr
	set cur_c cur
	set scur [llength $Q]
	if {$count == 0.0} {
		setcounter $nr 0 0.0 0.0 0.0
	} else {
		set percentstep [expr (1/$count) * 100.0]
		set total [expr $count / $fact]
		setcounter $nr 0 $total 0.0 $total
		for {set i 0} {$i<= $count} {incr i 100} {
			lappend afters [after $i setcounter $nr [expr $i*$percentstep] \
			$total [expr $i / $fact] [expr ($count - $i) / $fact]]
		}
		lappend afters [after $i setcounter $nr 100  $total $total 0.0]
	}
}

proc freeze {id count} {
	upvar #0 S_[set id](nr) nr
	
	cancel $nr
	enq $id freeze $count
}

proc stop {id count} {
	upvar #0 S_[set id](nr) nr

	cancel $nr
	stopcolor $nr
	enq $id stop $count
}

proc cancel {nr} {
	upvar #0 W_[set nr]_afters afters
	if {[info exists afters]} {
		foreach a $afters {
			after cancel $a
		}
	}
	catch {unset afters}
}

proc enq {id cmd count} {
	upvar #0 G_[set id](cur_n) cur_n		;# percentage (used as integer, _not_ float!)
	upvar #0 G_[set id](cur_t) cur_t		;# total  (float)
	upvar #0 G_[set id](cur_d) cur_d		;# done  (float)
	upvar #0 G_[set id](cur_r) cur_r		;# remains  (float)
	upvar #0 G_[set id](cur_c) cur_c		;# command (cur,freeze,stop)
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](Q) Q
	upvar #0 G_[set id](ofid) ofid

	if {[string compare $cmd stop] == 0 || [string compare $cmd freeze] == 0} {
		if {$count > $cur_t} {
			warn $ofid "rescaling $cmd argument > last wait count ($count > $cur_t)"
			set count $cur_t
		} 
		if {$count > $cur_r} {
			warn $ofid "ignoring $cmd argument > remains count ($count > $cur_r)"
			set count ""
		}
	}
	if {[string compare $cur_c cur] == 0} {
		set cur_c $cmd
		if {[string compare $count ""] != 0 && $cur_t != 0} {
			set cur_r $count
			set cur_d [expr $cur_t - $cur_r]
			set cur_n	[expr int(($cur_d / $cur_t) * 100.0)]
		}
		lappend Q [list $cur_c $cur_n $cur_t $cur_d $cur_r]
		debug 1 "enq $cur_c $cur_n $cur_t $cur_d $cur_r"
	} else {
		debug 1 "enq skip: $cur_c $cur_n $cur_t $cur_d $cur_r"
	}
}

proc showcur {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](step) step
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](Q) Q
	upvar #0 G_[set id](cur_n) cur_n
	upvar #0 G_[set id](cur_t) cur_t
	upvar #0 G_[set id](cur_d) cur_d
	upvar #0 G_[set id](cur_r) cur_r
	upvar #0 G_[set id](cur_c) cur_c

	set Qlen [llength $Q]
	debug 1 "showcur $nr $scur Qlen=$Qlen"
	if {$scur == $Qlen} {
		debug 1 "	showcur $nr $scur==Qlen $cur_c $cur_n $cur_t $cur_d $cur_r"
		showcounter $nr $cur_c $cur_n $cur_t $cur_d $cur_r
		set step $scur
	} else {
	 	set e [lindex $Q $scur]
		debug 1 "	showcur $nr $scur $e"
		eval showcounter $nr $e
		set step $scur
	}
}

proc showq {nr tell cur {prev ""}} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](Q) Q
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](sprev) sprev
	upvar #0 G_[set id](oldscur) oldscur
	upvar #0 G_[set id](oldsprev) oldsprev
	upvar #0 S_[set id](nr) w_nr
	upvar #0 G_[set id](cur_c) cur_c

	set last [llength $Q]
	debug 1 "showq ($nr)($tell)($cur)($prev)last=$last"
	if {[string compare $prev ""] == 0} {
		set prev [expr $cur - 1]
	}
	if {$cur >= 0  && ($cur < $last || ([string compare $cur_c cur] == 0 && $cur == $last))} {
		set oldscur $scur
		set oldsprev $sprev
		if {$prev >= 0 && $prev < $last} {
			set sprev $prev
		} else {
			set sprev -1
		}
		set scur $cur
		foreach nr $w_nr {
			showcur $nr
		}
		if {$tell} {
			mcastsend $id $cur
		}
	}
}

proc parse {id s} {
	upvar #0 G_[set id](ofid) ofid
	set line $s
	global debug
	global quit
	debug 4 "llength line: [llength $line]"
	debug 4 "line(0): [lindex $line 0]"
	debug 4 "line(1): [lindex $line 1]"
	switch -exact -- [string tolower [lindex $line 0]] {
	 	  "" {
			# skip
		} wait {
			set count [lindex $line 1]
			# debug 1 "aniwait $count"
			if {[llength $line] != 2} {
				warn $ofid "wrong number of arguments ([expr [llength $line] - 1]); usage: wait nr"
			} elseif {[string compare $count ""] == 0} {
				warn $ofid "wait needs argument"
			} elseif {![isdouble $count]} {
				warn $ofid "invalid wait argument, no float number: $count"
			} else {
				countdown $id $count
			}
		} stop {
			set count [lindex $line 1]
			if {[string compare $count ""] == 0} {
			} elseif {![isdouble $count]} {
				warn $ofid "invalid stop argument, no float number: $count"
			}
			stop $id $count
		} freeze {
			set count [lindex $line 1]
			if {[string compare $count ""] == 0} {
			} elseif {![isdouble $count]} {
				warn $ofid "invalid freeze argument, no float number: $count"
			}
			freeze $id $count
		} debuglvl {
			output stdout "\tdebuglvl $debug"
			debug 4 "llength: [llength $line]"
			debug 4 "1: [lindex $line 1]"
			if {([llength $line] >= 2) && ([string compare [lindex $line 1] ""] != 0)} {
				set debug [lindex $line 1]
			}
		} help {
			output stdout "cmdinterp commands:"
			output stdout "\tdebuglvl \[nr\]"
			output stdout "\twait nr"
			output stdout "\tstop"
			output stdout "\tfreeze"
			output stdout "\thelp"
		} default {
			warn $ofid "cmdinterp: unknown command: $line"
		}
	}
	return bla ;# XXXX
}



proc cmdinterp { id line chan argl } {
	upvar #0 G_[set id](Q) Q
	upvar #0 S_[set id](nr) w_nr

	debug 1 "cmdinterp ($line) ($argl)"
	set g [lindex $argl 0]
	set idx [llength $Q]
	# lappend Q [parse $id [string trim $line]]
	parse $id [string trim $line]
	showq [lindex $w_nr 0] 0 $idx [expr $idx -1]
	return 1
}

proc cmdcleanup { id chan argl } {
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) w_nr

	if {[string compare $ifid $chan] == 0} {
		set ieof 1
		foreach nr $w_nr {
			update_reuse $nr
		}
	}
}


proc mcastconnect {mcast ofid } {
	set cfid {}
	set fields [split $mcast "!"]
#puts stderr "mcastconnect $fields"
	if {([llength $fields] == 3) &&
		([string compare [lindex $fields 0] tcp] == 0)} {
		set host [lindex $fields 1]
		set port [lindex $fields 2]
		if {[catch {socket $host $port} foo]} {
			warn $ofid "cannot connect to mcast $mcast: $foo"
		} else {
			warn $ofid "connected to mcast $mcast: $foo"
			set cfid $foo
		}
	}
	return $cfid
}

proc mcastinterp { id line chan argl } {
	upvar #0 G_[set id](Q) Q
	upvar #0 S_[set id](nr) w_nr

	debug 1 "mcastinterp ($line) ($argl)"
	set g [lindex $argl 0]
	if {[isinteger  [lindex $line 0]]} {
		showq [lindex $w_nr 0] 0 [lindex $line 0]
	}
	return 1
}

proc mcastcleanup { id chan argl } {
	upvar #0 G_[set id](cfid) cfid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) nr

	if {[string compare $cfid $chan] == 0} {
		set cfid ""
	}
}

proc mcastsend {id nr} {
	upvar #0 G_[set id](cfid) cfid

# puts stderr "mcastsend $id $nr $cfid"
	if {[string compare $cfid ""] != 0} {
		if {[catch {output $cfid $nr} msg]} {
			debug 1 "mcastsend $nr -> $msg"
		} else {
			debug 1 "mcastsend $nr OK?"
		}
	}
}

proc debug { lvl s } {
	global debug
	if {$debug >= $lvl } {
		catch {
			puts stderr "DEBUG: aniwaitsrv: $s"
			flush stderr
		}
	}
}
proc warn { fid s } {
	global PID
	catch {
		puts $fid "aniwaitsrv:$PID warning: $s"
		flush $fid
	}
}
proc output { c s } {
	if {[catch {
		puts $c $s
		flush $c
	} msg]} {
		catch {puts stderr "output: $msg"}
	}
}

proc msg { fid s } {
	global PID
	catch {puts $fid "aniwaitsrv:$PID $s" }
}


proc handle { chan info hproc cproc id args } {
	fileevent $chan readable {}
	set read [gets $chan line]
	debug 2 "handle $info $hproc $cproc read: $read"
	if {$read < 0} {
		if {[eof $chan]} {
			debug 1 "Connection closed by client $info."
			 $cproc $id $chan $args
		} elseif {[fblocked $chan]} {
			warn stderr "handle $info $hproc $cproc blocked: $line"
			fileevent $chan readable \
				[concat handle $chan $info $hproc $cproc $id $args ]
		} else {
			warn stderr "handle $info error?: $line"
			$cproc $id $chan $args
		}
	} else {
		debug 3 "handle $info: $line"
		set res [$hproc $id $line $chan $args] 
		if {$res > 0} {
			fileevent $chan readable \
				[concat handle $chan $info $hproc $cproc $id $args]
		} elseif {$res < 0} {
			catch {close $chan}
			$cproc $id $chan $args
		}
	}

}

proc quit {nr} {
	global B

	msg stderr "exiting"
	if {([string compare $B(interface,name) ""] != 0) &&
	    $B(interface,cleanup) &&
	    [file readable $B(interface,name)]} {
		catch {file delete -force $B(interface,name)}
	}
	if {([string compare $B(pidsfile,name) ""] != 0) &&
	    $B(pidsfile,cleanup)} {
		catch {file delete -force $B(pidsfile,name)}
	}
	exit
}
proc mydestroy {nr} {
	upvar #0 W_[set nr](busy) busy
	upvar #0 W_[set nr](destroying) destroying

	debug 1  "mydestroy $nr {"

	set destroying 1
	if {$busy} {
		debug 1  "mydestroy $nr waiting for $busy"
		trace variable busy wu "destroy_when_not_busy $nr"
	} else {
		destroy_window_and_contents $nr
	}
	debug 1  "mydestroy $nr }"
}
proc destroy_when_not_busy {nr args} {
	debug 1  "destroy_when_not_busy $nr $args {"
	destroy_window_and_contents $nr
	debug 1  "destroy_when_not_busy $nr $args }"
}
proc destroy_window_and_contents {nr} {
	global windows
	upvar #0 W_[set nr](t) t

	debug 1  "destroy_window_and_contents $nr {"
	destroy $t
	set windows [ldelete $windows $t]
	destroy_contents $nr
	cleanupnr $nr
	if {[llength $windows] <= 0} {
		quit $nr
	}
	debug 1  "destroy_window_and_contents $nr }"
}
proc destroy_contents {nr} {
	debug 1  "destroy_contents $nr {"
	cleanupid $nr
	debug 1  "destroy_contents $nr }"
}
proc cleanupid {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 S_[set id](nr) w_nr

	debug 1 "cleanup $nr {"
	set w_nr [ldelete $w_nr $nr]
	if {[llength $w_nr] == 0} {
		upvar #0 G_[set id](cfid) cfid
		upvar #0 G_[set id](ifid) ifid
		global reuseQ

		debug 1 "cleanup $nr : id $id cfid $cfid ifid $ifid"
		if {[info exists cfid] && ([string compare $cfid ""] != 0)} {
			fileevent $cfid readable {}
			close $cfid
			set cfid ""
		}
		if {[info exists ifid] && ([string compare $ifid ""] != 0)} {
			fileevent $ifid readable {}
			close $ifid
			set ifid ""
		}
		catch {unset G_[set id]}
		catch {unset G_[set id]_name2edge}
		catch {unset G_[set id]_col}
	}
	debug 1 "cleanup $nr }"
}
proc cleanupnr {nr} {
	global reuseQ
	upvar #0 W_[set nr](id) id
	upvar #0 S_[set id](nr) w_nrs

	debug 1 "cleanupnr $nr: reuseQ ($reuseQ) {"
	set w_nrs [ldelete $w_nrs $nr]
	set reuseQ [ldelete $reuseQ $nr]
	catch {unset W_[set nr]}
	if {[llength $w_nrs] <= 0} {
		catch {unset S_[set id]}
	}
	
	debug 1 "cleanup $nr -> reuseQ ($reuseQ) }"
}

proc cleanupchan { id chan } {
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) nr

	debug 1  "cleanupchan $id $chan {"
	if {[string compare $ifid $chan] == 0} {
		set ieof 1
		update_reuse $nr
	}
	debug 1  "cleanupchan $id $chan }"
}

proc setup_B {defAddr defPort fname host} {
	global B

	set B(interface,name) $fname
	set B(pidsfile,name) "$fname.pid"
	set B(statusfile,name) "$fname.stat"
	set B(interface,cleanup) 0
	set B(pidsfile,cleanup) 0
	set pid [pid]

	if {[try_connect $B(interface,name) 0]} {
		warn stderr "not replacing running server (aborting)"
		exit 0
	} else {
		# warn stderr "could not connect to running server"
	}
	if {[catch {open $B(statusfile,name) a} fid]} {
		warn stderr "cannot open (append) $B(statusfile,name)"
		exit 1
	} else {
		puts $fid "starting $host $pid"
		close $fid
	}
	if {[catch {open $B(statusfile,name) r} fid]} {
		warn stderr "cannot open (read) $B(statusfile,name)"
		exit 1
	} elseif {[gets $fid line] < 0} {
		warn stderr "cannot read $B(statusfile,name)"
		exit 1
	} else {
		close $fid
		set words [split $line]
		set stat [lindex $words 0]
		set h [lindex $words 1]
		set p [lindex $words 2]
		if {[string compare $stat starting] == 0 &&
		    [string compare $host $h] == 0 &&
		    [string compare $pid $p] != 0} {
			set found 0
			if {[catch {open "|ps -ef"} pidsfd]} {
				warn stderr "could not run: ps -ef"
				exit 1
			} else {
				while {[gets $pidsfd line] >= 0} {
					debug 25 "$pid: looking for $p in: $line"
					if {[lindex $line 1] == $p}  {
						msg stderr "found running server pid: $p"
						set found 1
						break
					}
				}
				close $pidsfd
			}
			if {$found} {
				set i 0
				set maxtries 10
				while {$i < $maxtries} {
					if {$i == ($maxtries - 1)} {
						set sock [try_connect $B(interface,name) 1]
					} else {
						set sock [try_connect $B(interface,name) 0]
					}
					if {$sock} {
						warn stderr "attempt $i: not replacing running server (aborting)"
						exit 0
					} elseif {$i == ($maxtries - 1)} {
						warn stderr "attempt $i: could not connect to running server"
					}
					incr i
				}
			} else {
				msg stderr "did not find running server pid: $p"
			}
		}
	}
	if {[try_connect $B(interface,name) 1]} {
		warn stderr "not replacing running server (aborting)"
		exit 0
	} else {
		warn stderr "could not connect to running server"
	}
	if [catch {open $B(interface,name) w} fid] {
		warn stderr "cannot create $B(interface,name): $fid"
		return 0
	}
	set B(interface,cleanup) 1

	msg stderr "trying to listen on tcp!$defAddr!$defPort"
	if {$defAddr == {}} {
			set c [socket -server Baccept $defPort]
		} else {
			set c [socket -server Baccept -myaddr $defAddr $defPort]
		}
	set cfg [fconfigure $c -sockname]
	set ipNr [lindex $cfg 0]
	set hostName [lindex $cfg 1]
	set portNr [lindex $cfg 2]
	msg stderr "listening on tcp!$ipNr!$portNr"
	
	puts $fid "$ipNr $hostName $portNr"
	close $fid
	
	if {![catch {open "$B(pidsfile,name)" r} pfid]} {
		while {[gets $pfid line] >= 0} {
			set p [string trim $line]
			msg stderr "cleaning up: kill $p"
			catch {exec kill $p}
			msg stderr "cleaning up: kill -9 $p"
			catch {exec kill -9 $p}
		}
		catch {close $pfid}
	}
	catch {file delete -force $B(pidsfile,name)}
	if {[catch {open "$B(pidsfile,name)" w} pfid]} {
		warn stderr "cannot open $B(pidsfile,name) for writing: $pfid"
	} else {
		set B(pidsfile,cleanup) 1
		puts $pfid $pid
		catch {close $pfid}
	}
	catch {file delete -force $B(statusfile,name)}
	return 1
}

proc Baccept {socket remoteHost remotePort} {
	global B
	
    set nSock      $socket
 	set B(busy,$socket) 0
	fconfigure $socket -buffering line -blocking 0
	fileevent $socket readable [list handle $socket <B> Binterp Bcleanup ign_id]
	msg stderr "Connection accepted from client tcp!$remoteHost!$remotePort"
}  


proc Binterp { ign_id line chan argl } {
	return [continue_reading_after_process_B_line $line $chan]
}

proc Bcleanup { ign_id chan argl } {
	global B

	if {![catch {fconfigure $chan -peername} cfg]} {
		set remoteHost [lindex $cfg 0]
		set remotePort [lindex $cfg 2]
		msg stderr "Connection closed by client tcp!$remoteHost!$remotePort"
	}
	catch {close $chan}
	unset B(busy,$chan)
}

proc continue_reading_after_process_B_line { line fid } {
	global B
	set name [string trim $line]
	set result 1
	if {([string compare $name -exit] == 0) ||
	    (([string compare [lindex $name 0] -args] == 0) &&
	     ([string compare [lindex $name 1] -exit] == 0))} {
		# exit will do this for us: cleanupchan
		quit ignored
	} elseif {[string compare [lindex $name 0] -eof] == 0} {
		debug 1 "B eof"
		trace variable B(busy,$fid) w "close_B_when_ready $fid"
		return 0
	} elseif {[string compare [lindex $name 0] -debug] == 0} {
		debug 1 "B debug: [lrange $name 1 end]"
	} elseif {[string compare [lindex $name 0] -args] == 0} {
		debug 1 "B debug: [lrange $name 1 end]"
		set result [continue_reading_after_process_args [lrange $name 1 end] $fid $fid]
	} else {
		debug 1 "B debug: unknown line: $line"
	}
	return $result
}

proc close_B_when_ready {fid varName index op} {
	global B
	debug 1 "close_B_when_ready ($fid)($varName)($index)($op)"
	switch $op {
		w {
			if {[string compare "busy,$fid" $index] == 0} {
				if {$B(busy,$fid) == 0} {
					debug 1 "close_B_when_ready: closing fid (B(busy,$fid) == 0)"
					if {![catch {fconfigure $fid -peername} cfg]} {
						set remoteHost [lindex $cfg 0]
						set remotePort [lindex $cfg 2]
						msg stderr "Closing connection to client tcp!$remoteHost!$remotePort"
					}
					close $fid
					unset B(busy,$fid)
				} else {
					debug 1 "close_B_when_ready: not yet closing fid (B(busy,$fid) == $B(busy,$fid))"
				}
			} else {
					debug 1 "close_B_when_ready: unrelated update (index=$index)"
				
			}
		}
	}
}

proc continue_reading_after_process_args {argv ifid ofid {name stdin} } {
	global env
	global id
	
	set reuse 0
	set files {}
	set result 1
	set mcast ""
	set title "Aniwait"
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast $env(TORXMCASTID)
	}

	set args $argv
	
	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse 1
		  } -m {
				if {[llength $args] <= 0} {
		  			warn $ofid "missing argument for flag: $a"
				} else {
					set mcast [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warn $ofid "missing argument for flag: $a"
				} else {
					set title "Aniwait"
					append title " - "
					append title [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -* {
		  		warn $ofid "unknown flag: $a"
				set result -1
		  } default {
		  		lappend files $a
		  }
		}
	}

	if {($result == -1) || ($ifid == {}) || ([llength $files] != 0)} {
		warn $ofid "usage: aniwait \[-r\] \[-m mcastid\] \[-t title\]"
		set result -1
	} else {
		set cfid [mcastconnect $mcast $ofid]
		incr id
		start $id $reuse $ifid $ofid $cfid $name $title
		set result 0
	}
	return $result
}


proc start {id rse aifid aofid acfid name atitle} {
	global idinfo
	upvar #0 G_[set id](scur) scur
	upvar #0 G_[set id](sprev) sprev
	upvar #0 G_[set id](oldscur) oldscur
	upvar #0 G_[set id](oldsprev) oldsprev
	upvar #0 G_[set id](Q) Q
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 G_[set id](iname) iname
	upvar #0 G_[set id](ofid) ofid
	upvar #0 G_[set id](cfid) cfid
	upvar #0 G_[set id](busy) busy
	upvar #0 G_[set id](destroying) destroying
	upvar #0 G_[set id](reuse) reuse
	upvar #0 G_[set id](cur_c) cur_c
	upvar #0 G_[set id](title) title
	global env

	set reuse $rse
	set ifid $aifid
	set ieof 0
	set iname $name
	set ofid $aofid
	set cfid $acfid
	set title $atitle

	# we use busy and destroying to synchronise with the destroy
	# to avoid trying to write to a window that is just destroyed
	set busy 0
	set destroying 0

	set idinfo($id,started) 1
	set idinfo($id,closed) 0

	if {[info exists B(busy,$ofid)]} {
		incr B(busy,$ofid)
	}
	
	if {[string compare $ifid stdin ] != 0} {
		fconfigure $ifid -buffering line -blocking 0
	}

	set Q {}
	set scur "0"
	set sprev ""
	set oldscur ""
	set oldsprev ""

	set nr [getcanvas $id $reuse $ofid]
	setcounter $nr 0 0.0 0.0 0.0
	set cur_c cur
	enq $id freeze 0.0

	fileevent $ifid readable [list handle $ifid $name cmdinterp cmdcleanup $id]
	debug 1 "start: started cmdinterp $id $ifid $name"
	if {[string compare $cfid ""] != 0} {
		fileevent $cfid readable [list handle $cfid "<mcast>" mcastinterp mcastcleanup $id ]
	}

}

# =================================

proc find_user {host} {
	global env

	if {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set fname [file join $dir "$pfx-$user-$display"]
	regsub -all : $fname _ fname
	return $fname
}

proc try_connect { fname do_warn } {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			set words [split $line]
			set ipNr [lindex $words 0]
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			if {![catch {socket $ipNr $portNr} socket]} {
				return 1
			} else {
				warn stderr "could not connect tcp!$ipNr!$portNr: $socket"
			}
		}
		close $fid
	} elseif {$do_warn} {
		warn stderr "$fid"
	}
	return 0
}

proc main {} {
	global argv
	
	image create bitmap up -data {
#define up_width 5
#define up_height 3
static unsigned char up_bits[] = {
   0x04, 0x0e, 0x1f};
	}
	image create bitmap down -data {
#define down_width 5
#define down_height 3
static unsigned char down_bits[] = {
   0x1f, 0x0e, 0x04};
	}

	if {[llength $argv] > 0} {
		set s [join $argv " "]
		warn stderr "aniwaitsrv: ignoring arguments: $s"
		warn stderr "usage: aniwaitsrv"
	}
	set host [exec uname -n]
	set display [find_display $host]
	set user [find_user $host]
	set fname [mk_filename /tmp aniwait $user $display]

	if {![setup_B 127.0.0.1 0 $fname $host]} {
		exit 1
	}
	
#	if {[continue_reading_after_process_args $argv stdin stderr] < 0} {
#		# hmmm... is this what we want,
#		# or should server remain running?
#		quit ignored
#	}
}

main

vwait quit
exit
