#! @PERL@

$debug = 0;

$| = 1; # do not buffer!

require 5.002;
use Getopt::Std;

%opts = ();
$result = getopts('acCnNrsStT', \%opts);

$mode_count = 1;
$mode_stepnrs = 0;
$mode_super = 0;
$mode_r = 0;

if ($opts{'a'}) {
	$mode_count = 1;
	$mode_stepnrs = 0;
	$mode_super = 1;
}
if ($opts{'c'}) {
	$mode_count = 1;
}
if ($opts{'C'}) {
	$mode_count = 0;
}
if ($opts{'n'}) {
	$mode_stepnrs = 1;
}
if ($opts{'N'}) {
	$mode_stepnrs = 0;
}
if ($opts{'r'}) {
	$mode_r = 1;
}
if ($opts{'s'}) {
	$mode_super = 1;
}
if ($opts{'S'}) {
	$mode_super = 0;
}
if ($opts{'t'}) {
	$mode_count = 0;
	$mode_stepnrs = 1;
	$mode_super = 0;
}
if ($opts{'T'}) {
	$mode_count = 0;
	$mode_stepnrs = 1;
	$mode_super = 1;
}


initialize();

while(<>) {
	chomp;
	/^ABSTRACT/ && do handle_action($_ , 0);
	/^STATEID/ && do handle_stateid($_ , 0);
}

finalize();

sub initialize {
	$nostateids = 0;
}

sub handle_action {
	my($line, $expected) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr, $iokind_chan_pco, $suspension, $action,
	   $iokind,$chan_pco_opt, $chan_pco, $chan, $pco, $inst);
	$ignore = shift(@words); # shift out ABSTRACT
	$stepnr = shift(@words);
	$iokind_chan_pco = shift(@words);
	$suspension = pop(@words);
	$suspension = strip_parentheses($suspension);
	$action = join(" ", @words);
	$action = strip_parentheses($action);
	if ($theSource eq '' ) { # no STATEID
		$nostateids = 1;
		startgraph(0, $stepnr - 1);
	}
	if ($iokind_chan_pco =~ /^(input|output)(\(([^\)]*)\))?$/) {
		$iokind=$1;
		$chan_pco_opt = $2;
		$chan_pco = $3;
		if ($chan_pco =~ /^([^:]+):([^:]+)$/) {
			$chan = $1;
			$pco = $2;
			add_nonsuspension($stepnr, $iokind, $action );
		} elsif ($chan_pco =~ /^([^:]+)$/) {
			$chan = $1;
			$pco = '';
			$inst = '('.$chan.')';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind);
			} else {
				msg("warning: no pco for non-suspension action nr: $stepnr");
				add_nonsuspension($stepnr, $iokind, $action );
			}
		} else {
			$chan = '';
			$pco = '';
			$inst = 'tester';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind);
			} else {
				add_nonsuspension($stepnr, $iokind, $action );
			}
		}
		if ($nostateids) { # no STATEID
			extendgraph(0, $stepnr);
		}
	} else {
		msg("parse error: could not parse iokind/channel/pco for action nr: $stepnr");
	}
}

sub add_suspension {
	my($stepnr, $iokind) = @_;
	print STDERR "add_suspension ($stepnr, $iokind)" if $debug;
	if ($iokind eq "input") {
		$theAction='suspension?';
	} elsif ($iokind eq "output") {
		$theAction='quiescence!';
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}
sub add_nonsuspension {
	my($stepnr, $iokind, $action) = @_;
	my($i, $iut, $actpfx);
	print STDERR "add_nonsuspension ($stepnr, $iokind, $action)" if $debug;
	if ($iokind eq "input") {
		$theAction=$action;
	} elsif ($iokind eq "output") {
		$theAction=$action;
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}

# STATEID 0 SPEC super 0 init 0
sub handle_stateid {
	my($line, $expected) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr, $role, $supe, $k);
	$ignore = shift(@words); # shift out ABSTRACT
	$stepnr = shift(@words);
	$role = shift(@words);
	if ($role ne "SPEC") {
		return;
	}
	$ignore = shift(@words); # shift out super kw (should check)
	$super = shift(@words);
	# ignore the rest
	if ($theSource ne '' && $theAction ne '') {
		extendgraph($super, $stepnr);
	} else {
		startgraph($super, $stepnr);
	}
}

sub startgraph {
	my($super, $step) = @_;

	$theInit = mkstate($super, $step);
	$theSource = $theInit;
}
sub extendgraph {
	my($super, $step) = @_;

	$theDest = mkstate($super, $step);
	$k = join($;, ($theSource, $theAction, $theDest));
	$trans{$k}++;
	$state{$theSource} = 1;
	$state{$theDest} = 1;
	$theSource = $theDest;
}

sub finalize {
	my($k, $s, $a, $d, $c, $t, $nstates, $ntrans);
	$nstates =  keys(%state);
	$ntrans = keys(%trans);
	if ($theInit eq '') {
		return;
	}
	print "des ($theInit, $nstates, $ntrans)\n";
	foreach $k (sort (bystate keys(%trans))) {
		($s, $a, $d) = split($;, $k);
		if ($mode_count) {
			$c = $trans{$k};
			$t = $a.'@'.$c;
		} else {
			$t = $a;
		}
		print "($s, \"$t\", $d)\n";
	}
}

sub mkstate {
	my($super, $step) = @_;
	if ($mode_stepnrs && $mode_super) {
		return $super.'_'.$step;
	} elsif ($mode_stepnrs) {
		return $step;
	} elsif ($super) {
		return $super;
	} else {
		return $step;
	}
}


sub msg {
	my($s) = @_;
	print STDERR "log2msc: ".$s."\n";
}

sub strip_parentheses {
	my($s) = @_;
	if ($s =~ /^\((.*)\)$/) {
		return $1;
	} else {
		return $s;
	}
}

sub strip_quotes {
	my($s) = @_;
	if ($s =~ /^"(.*)"$/) {
		return $1;
	} else {
		return $s;
	}
}

sub bystate {
	my($sa, $aa, $da, $sb, $ab, $db, $ssa, $sna, $ssb, $snb);
	($sa, $aa, $da) = split($;, $a);
	($sb, $ab, $db) = split($;, $b);
	if ($sa =~ /_/) {
		($ssa, $sna) = split('_', $sa);
		($ssb, $snb) = split('_', $sb);
		($dsa, $dna) = split('_', $da);
		($dsb, $dnb) = split('_', $db);
		if ($mode_r) {
			if ($sna != $snb) {
				return $sna <=> $snb;
			} elsif ($ssa != $ssb) {
				return $ssa <=> $ssb;
			} elsif ($dna != $dnb) {
				return $dna <=> $dnb;
			} else {
				return $dsa <=> $dsb;
			}
		} else {
			if ($ssa != $ssb) {
				return $ssa <=> $ssb;
			} elsif ($sna != $snb) {
				return $sna <=> $snb;
			} elsif ($dsa != $dsb) {
				return $dsa <=> $dsb;
			} else {
				return $dna <=> $dnb;
			}
		}
	} else {
		if ($sa != $sb) {
			return $sa <=> $sb;
		} else {
			return $da <=> $db;
		}
	}
}
