#! @PERL@

$debug = 0;

$| = 1; # do not buffer!


initialize();

loop:
while(<>) {
	chomp();
	/^CONFIGEND/ && do handle_config_end($_);
	/^CONFIG/    && do handle_config($_);
	/^ABSTRACT/  && do handle_action($_ , '', '');
	/^EXPECTED/  && do handle_action($_, 'Expected: ', '');
	/^REASON/   && do handle_reason($_);
	/^REFUSED/   && do handle_action($_, '(Refusal: ', ')');
	/^EOF/	     && last loop;
}

finalize();
exit;

sub initialize {
	%instances = ();
	$config_end_handled = 0;
}

sub handle_config_end {
	my($line) = @_;
	print STDERR "line: ($line) " if $debug;
	$config_end_handled = 1;
	if ($cfgfilename ne "") {
		$title = $cfgfilename;
		if (1 && $primer ne "") {
			$title .= '-'.$primer;
		}
		if (1 && $mutant ne "") {
			$title .= '-'.$mutant;
		}
		print STDERR "msc ($title) " if $debug;
		$result = 'msc ';
		$result .= $title.';';
		$result .= "\n";
		print $result;
	}
}

sub handle_config {
	my($line) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $keyword);
	print STDERR "words: (@words) " if $debug;
	$ignore = shift(@words); # shift out CONFIG
	$keyword = shift(@words); # shift out config keyword
	if ($keyword eq "CONFIGFILE") {
		$cfgfilename = shift(@words);
		$cfgfilename =~ s/\.if$//;
	} elsif ($keyword eq "PRIMER") {
		$primer = shift(@words);
	} elsif ($keyword eq "MUTANT") {
		$mutant = shift(@words);
	}
}

sub handle_reason {
	if (! $config_end_handled) {
		handle_config_end("");
	}
	my($line) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr);
	$ignore = shift(@words); # shift out ABSTRACT
	$stepnr = shift(@words);
	$reason{$stepnr} = join(" ", @words);
}

sub handle_action {
	if (! $config_end_handled) {
		handle_config_end("");
	}
	my($line, $actpfx, $actsfx) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr, $iokind_chan_pco, $suspension, $action,
	   $iokind,$chan_pco_opt, $chan_pco, $chan, $pco, $inst);
	$ignore = shift(@words); # shift out ABSTRACT
	$stepnr = shift(@words);
	$iokind_chan_pco = shift(@words);
	$suspension = pop(@words);
	$suspension = strip_parentheses($suspension);
	$action = join(" ", @words);
	$action = strip_parentheses($action);
	if (defined $reason{$stepnr} && $reason{$stepnr} ne "") {
		if ($actpfx =~ /:/) {
			$actpfx =~ s/:/ ($reason{$stepnr}):/;
		} else {
			$actpfx .= $reason{$stepnr}.': ';
		}
	}
	if ($iokind_chan_pco =~ /^(input|output)(\(([^\)]*)\))?$/) {
		$iokind=$1;
		$chan_pco_opt = $2;
		$chan_pco = $3;
		if ($chan_pco =~ /^([^:]+):([^:]+)$/) {
			$chan = $1;
			$pco = $2;
			$inst = $pco;
			add_nonsuspension($stepnr, $iokind, $inst, $actpfx, $actsfx, $action );
		} elsif ($chan_pco =~ /^([^:]+)$/) {
			$chan = $1;
			$pco = '';
			$inst = '('.$chan.')';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind, $inst, $actpfx, $actsfx, $action);
			} else {
				msg("warning: no pco for non-suspension action nr: $stepnr");
				add_nonsuspension($stepnr, $iokind, $inst, $actpfx, $actsfx, $action );
			}
		} else {
			$chan = '';
			$pco = '';
			$inst = 'tester';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind, $inst, $actpfx, $actsfx, $action);
			} else {
				add_nonsuspension($stepnr, $iokind, $inst, $actpfx, $actsfx, $action );
			}
		}
	} else {
		msg("parse error: could not parse iokind/channel/pco for action nr: $stepnr");
	}
}

sub add_instance {
	my($iname) = @_;
	my($i);
	if (! defined($instances{$iname})) {
		definstance($iname);
		$i = instance($iname);
		print "inst $i;\n";
		print "$i: instancehead \"$iname\";\n";
	}
}


sub definstance {
	my($iname) = @_;
	my($l, $i, $normalize);
	$normalize = 1;
	if ($normalize) {
		$l = scalar keys %instances;
		$instances{$iname} = "i$l".'_';
		if ($iname eq 'iut') {
			# nothing special to add
		} elsif ($iname =~ /^\(/) {
			$instances{$iname} .= 'chan';
		} else {
			$instances{$iname} .= 'pco';
		}
		$instances{$iname} .= '_'.strip_parentheses($iname);
	} else {
		$instances{$iname} = "$iname";
	}
}

sub instance {
	my($iname) = @_;
	add_instance($iname);
	return $instances{$iname};
}

sub add_suspension {
	my($stepnr, $iokind, $iname, $actpfx, $actsfx, $action) = @_;
	print STDERR "add_suspension ($stepnr, $iokind, $iname, $actpfx, $actsfx, $action)" if $debug;
	if ($iokind eq "input") {
		if ($action eq "") {
			add_nonsuspension($stepnr, $iokind, $iname, $actpfx, $actsfx, "(Suspension)");
		} elsif ($actpfx eq "") {
			add_nonsuspension($stepnr, $iokind, $iname, "(Refusal: ", ")", $action);
		} else {
			add_nonsuspension($stepnr, $iokind, $iname, $actpfx, $actsfx, $action);
		}
	} elsif ($iokind eq "output") {
		add_nonsuspension($stepnr, $iokind, $iname, $actpfx, $actsfx, "(Quiescence)");
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}
sub add_nonsuspension {
	my($stepnr, $iokind, $iname, $actpfx, $actsfx, $action) = @_;
	my($i, $iut);
	print STDERR "add_nonsuspension ($stepnr, $iokind, $iname, $actpfx, $actsfx, $action)" if $debug;
	# order is important for the ``instance'' statements, because
	# ``missing'' instances will be added from left to right
	# (and initially all instances are missing)
	$iut = instance("iut");
	$i = instance($iname);
	if ($iokind eq "input") {
		print "$i: out \"$actpfx$action$actsfx\" to $iut;\n";
		print "$iut: in \"$actpfx$action$actsfx\" from $i;\n";
	} elsif ($iokind eq "output") {
		print "$iut: out \"$actpfx$action$actsfx\" to $i;\n";
		print "$i: in \"$actpfx$action$actsfx\" from $iut;\n";
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}

sub finalize {
	my($i);
	foreach $i (values %instances) {
		print "$i: endinstance;\n";
	}
	print "endmsc;\n";
}


sub msg {
	my($s) = @_;
	print STDERR "log2msc: ".$s."\n";
}

sub strip_parentheses {
	my($s) = @_;
	if ($s =~ /^\((.*)\)$/) {
		return $1;
	} else {
		return $s;
	}
}

sub strip_quotes {
	my($s) = @_;
	if ($s =~ /^"(.*)"$/) {
		return $1;
	} else {
		return $s;
	}
}
